@@ -1,16 +1,32 @@
 #pragma once
 
+#include "wayfire/core.hpp"
+#include "wayfire/debug.hpp"
+#include "wayfire/geometry.hpp"
+#include "wayfire/opengl.hpp"
+#include "wayfire/region.hpp"
+#include "wayfire/scene-input.hpp"
+#include "wayfire/scene-operations.hpp"
+#include "wayfire/scene-render.hpp"
+#include "wayfire/scene.hpp"
+#include "wayfire/seat.hpp"
+#include "wayfire/signal-definitions.hpp"
+#include <memory>
 #include <wayfire/nonstd/reverse.hpp>
+#include <wayfire/plugins/common/util.hpp>
 #include <wayfire/plugins/wobbly/wobbly-signal.hpp>
 #include <wayfire/object.hpp>
 #include <wayfire/output-layout.hpp>
 #include <wayfire/nonstd/observer_ptr.h>
 #include <wayfire/render-manager.hpp>
-#include <wayfire/workspace-manager.hpp>
+#include <wayfire/signal-provider.hpp>
+#include <wayfire/workspace-set.hpp>
 #include <wayfire/view-transform.hpp>
 #include <wayfire/util/duration.hpp>
 #include <wayfire/util/log.hpp>
 #include <cmath>
+#include <wayfire/nonstd/wlroots-full.hpp>
+#include <wayfire/window-manager.hpp>
 
 
 namespace wf
@@ -43,7 +59,7 @@ namespace move_drag
  * when: Emitted output whenever the output where the drag happens changes,
  *   including when the drag begins.
  */
-struct drag_focus_output_signal : public signal_data_t
+struct drag_focus_output_signal
 {
     /** The output which was focused up to now, might be null. */
     wf::output_t *previous_focus_output;
@@ -57,7 +73,7 @@ struct drag_focus_output_signal : public signal_data_t
  * when: Emitted if snap-off is enabled and the view was moved more than the
  *   threshold.
  */
-struct snap_off_signal : public signal_data_t
+struct snap_off_signal
 {
     /** The output which is focused now. */
     wf::output_t *focus_output;
@@ -69,7 +85,7 @@ struct snap_off_signal : public signal_data_t
  * when: Emitted after the drag operation has ended, and if the view is unmapped
  *   while being dragged.
  */
-struct drag_done_signal : public signal_data_t
+struct drag_done_signal
 {
     /** The output where the view was dropped. */
     wf::output_t *focused_output;
@@ -80,7 +96,7 @@ struct drag_done_signal : public signal_data_t
     struct view_t
     {
         /** Dragged view. */
-        wayfire_view view;
+        wayfire_toplevel_view view;
 
         /**
          * The position relative to the view where the grab was.
@@ -93,7 +109,7 @@ struct drag_done_signal : public signal_data_t
     std::vector<view_t> all_views;
 
     /** The main view which was dragged. */
-    wayfire_view main_view;
+    wayfire_toplevel_view main_view;
 
     /**
      * The position of the input when the view was dropped.
@@ -136,7 +152,7 @@ inline static wf::pointf_t find_relative_grab(
  * It is primarily used to scale the view is a plugin needs it, and also to keep it
  * centered around the `grab_position`.
  */
-class scale_around_grab_t : public wf::view_transformer_t
+class scale_around_grab_t : public wf::scene::floating_inner_node_t
 {
   public:
     /**
@@ -159,23 +175,19 @@ class scale_around_grab_t : public wf::view_transformer_t
      */
     wf::point_t grab_position;
 
-    uint32_t get_z_order() override
+    scale_around_grab_t() : floating_inner_node_t(false)
+    {}
+
+    std::string stringify() const override
     {
-        return wf::TRANSFORMER_HIGHLEVEL - 1;
+        return "move-drag";
     }
 
-    wf::region_t transform_opaque_region(
-        wf::geometry_t box, wf::region_t region) override
+    wf::pointf_t scale_around_grab(wf::pointf_t point, double factor)
     {
-        // TODO: figure out a way to take opaque region into account
-        return {};
-    }
-
-    wf::pointf_t scale_around_grab(wf::geometry_t view, wf::pointf_t point,
-        double factor)
-    {
-        auto gx = view.x + view.width * relative_grab.x;
-        auto gy = view.y + view.height * relative_grab.y;
+        auto bbox = get_children_bounding_box();
+        auto gx   = bbox.x + bbox.width * relative_grab.x;
+        auto gy   = bbox.y + bbox.height * relative_grab.y;
 
         return {
             (point.x - gx) * factor + gx,
@@ -183,42 +195,57 @@ class scale_around_grab_t : public wf::view_transformer_t
         };
     }
 
-    wf::pointf_t transform_point(wf::geometry_t view, wf::pointf_t point) override
+    wf::pointf_t to_local(const wf::pointf_t& point) override
     {
-        LOGE("Unexpected transform_point() call for dragged overlay view!");
-        return scale_around_grab(view, point, 1.0 / scale_factor);
+        return scale_around_grab(point, scale_factor);
     }
 
-    wf::pointf_t untransform_point(wf::geometry_t view, wf::pointf_t point) override
+    wf::pointf_t to_global(const wf::pointf_t& point) override
     {
-        LOGE("Unexpected untransform_point() call for dragged overlay view!");
-        return scale_around_grab(view, point, scale_factor);
+        return scale_around_grab(point, 1.0 / scale_factor);
     }
 
-    wf::geometry_t get_bounding_box(wf::geometry_t view,
-        wf::geometry_t region) override
+    wf::geometry_t get_bounding_box() override
     {
-        int w = std::floor(view.width / scale_factor);
-        int h = std::floor(view.height / scale_factor);
-
-        auto bb = find_geometry_around({w, h}, grab_position, relative_grab);
-        return bb;
+        auto bbox = get_children_bounding_box();
+        int w     = std::floor(bbox.width / scale_factor);
+        int h     = std::floor(bbox.height / scale_factor);
+        return find_geometry_around({w, h}, grab_position, relative_grab);
     }
 
-    void render_with_damage(wf::texture_t src_tex, wlr_box src_box,
-        const wf::region_t& damage, const wf::framebuffer_t& target_fb) override
+    class render_instance_t :
+        public scene::transformer_render_instance_t<scale_around_grab_t>
     {
-        // Get target size
-        auto bbox = get_bounding_box(src_box, src_box);
+      public:
+        using transformer_render_instance_t::transformer_render_instance_t;
 
-        OpenGL::render_begin(target_fb);
-        for (auto& rect : damage)
+        void transform_damage_region(wf::region_t& region) override
         {
-            target_fb.logic_scissor(wlr_box_from_pixman_box(rect));
-            OpenGL::render_texture(src_tex, target_fb, bbox);
+            region |= self->get_bounding_box();
         }
 
-        OpenGL::render_end();
+        void render(const wf::render_target_t& target,
+            const wf::region_t& region) override
+        {
+            auto bbox = self->get_bounding_box();
+            auto tex  = this->get_texture(target.scale);
+
+            OpenGL::render_begin(target);
+            for (auto& rect : region)
+            {
+                target.logic_scissor(wlr_box_from_pixman_box(rect));
+                OpenGL::render_texture(tex, target, bbox);
+            }
+
+            OpenGL::render_end();
+        }
+    };
+
+    void gen_render_instances(std::vector<scene::render_instance_uptr>& instances,
+        scene::damage_callback push_damage, wf::output_t *shown_on) override
+    {
+        instances.push_back(std::make_unique<render_instance_t>(this,
+            push_damage, shown_on));
     }
 };
 
@@ -231,10 +258,10 @@ static const std::string move_drag_transformer = "move-drag-transformer";
 struct dragged_view_t
 {
     // The view being dragged
-    wayfire_view view;
+    wayfire_toplevel_view view;
 
     // Its transformer
-    nonstd::observer_ptr<scale_around_grab_t> transformer;
+    std::shared_ptr<scale_around_grab_t> transformer;
 
     // The last bounding box used for damage.
     // This is needed in case the view resizes or something like that, in which
@@ -242,7 +269,7 @@ struct dragged_view_t
     wf::geometry_t last_bbox;
 };
 
-inline wayfire_view get_toplevel(wayfire_view view)
+inline wayfire_toplevel_view get_toplevel(wayfire_toplevel_view view)
 {
     while (view->parent)
     {
@@ -252,10 +279,10 @@ inline wayfire_view get_toplevel(wayfire_view view)
     return view;
 }
 
-inline std::vector<wayfire_view> get_target_views(wayfire_view grabbed,
+inline std::vector<wayfire_toplevel_view> get_target_views(wayfire_toplevel_view grabbed,
     bool join_views)
 {
-    std::vector<wayfire_view> r = {grabbed};
+    std::vector<wayfire_toplevel_view> r = {grabbed};
     if (join_views)
     {
         r = grabbed->enumerate_views();
@@ -264,77 +291,99 @@ inline std::vector<wayfire_view> get_target_views(wayfire_view grabbed,
     return r;
 }
 
-/**
- * An object for storing per-output data.
- */
-class output_data_t : public noncopyable_t, public custom_data_t
+// A node to render the dragged views in global coordinates.
+// The assumption is that all nodes have a view transformer which transforms them to global (not output-local)
+// coordinates and thus we just need to schedule them for rendering.
+class dragged_view_node_t : public wf::scene::node_t
 {
+    std::vector<dragged_view_t> views;
+
   public:
-    output_data_t(wf::output_t *output, std::vector<dragged_view_t> views)
+    dragged_view_node_t(std::vector<dragged_view_t> views) : node_t(false)
     {
-        output->render->add_effect(&damage_overlay, OUTPUT_EFFECT_PRE);
-        output->render->add_effect(&render_overlay, OUTPUT_EFFECT_OVERLAY);
-
-        this->output = output;
-        this->views  = views;
+        this->views = views;
     }
 
-    ~output_data_t()
+    std::string stringify() const override
     {
-        output->render->rem_effect(&damage_overlay);
-        output->render->rem_effect(&render_overlay);
+        return "move-drag-view " + stringify_flags();
     }
 
-    void apply_damage()
+    void gen_render_instances(std::vector<scene::render_instance_uptr>& instances,
+        scene::damage_callback push_damage, wf::output_t *output = nullptr) override
     {
+        instances.push_back(std::make_unique<dragged_view_render_instance_t>(this, push_damage, output));
+    }
+
+    wf::geometry_t get_bounding_box() override
+    {
+        wf::region_t bounding;
         for (auto& view : views)
         {
             // Note: bbox will be in output layout coordinates now, since this is
             // how the transformer works
-            auto bbox = view.view->get_bounding_box();
-            bbox = bbox + -wf::origin(output->get_layout_geometry());
-
-            output->render->damage(bbox);
-            output->render->damage(view.last_bbox);
-
-            view.last_bbox = bbox;
+            auto bbox = view.view->get_transformed_node()->get_bounding_box();
+            bounding |= bbox;
         }
+
+        return wlr_box_from_pixman_box(bounding.get_extents());
     }
 
-  private:
-    wf::output_t *output;
-
-    std::vector<dragged_view_t> views;
-
-    // An effect hook for damaging the view on the current output.
-    //
-    // This is needed on a per-output basis in order to drive the scaling animation
-    // forward, if such an animation is running.
-    //
-    // TODO: We overdo damage, for ex. in the following cases:
-    // - Expo does not need any damage (can't really be fixed, since we don't know
-    // the plugin which uses this API).
-    // - If the view has not updated, and cursor has not moved
-    effect_hook_t damage_overlay = [=] ()
+    class dragged_view_render_instance_t : public wf::scene::render_instance_t
     {
-        apply_damage();
-    };
-
-    effect_hook_t render_overlay = [=] ()
-    {
-        auto fb = output->render->get_target_framebuffer();
-        fb.geometry = output->get_layout_geometry();
-
-        for (auto& view : wf::reverse(views))
+        wf::geometry_t last_bbox = {0, 0, 0, 0};
+        wf::scene::damage_callback push_damage;
+        std::vector<scene::render_instance_uptr> children;
+        wf::signal::connection_t<scene::node_damage_signal> on_node_damage =
+            [=] (scene::node_damage_signal *data)
         {
-            // Convert damage from output-local coordinates (last_bbox) to
-            // output-layout coords.
-            wf::region_t damage;
-            damage |= view.last_bbox + wf::origin(fb.geometry);
+            push_damage(data->region);
+        };
 
-            // Render the full view, always
-            // Not very efficient
-            view.view->render_transformed(fb, std::move(damage));
+      public:
+        dragged_view_render_instance_t(dragged_view_node_t *self, wf::scene::damage_callback push_damage,
+            wf::output_t *shown_on)
+        {
+            auto push_damage_child = [=] (wf::region_t child_damage)
+            {
+                push_damage(last_bbox);
+                last_bbox = self->get_bounding_box();
+                push_damage(last_bbox);
+            };
+
+            for (auto& view : self->views)
+            {
+                auto node = view.view->get_transformed_node();
+                node->gen_render_instances(children, push_damage_child, shown_on);
+            }
+        }
+
+        void schedule_instructions(std::vector<scene::render_instruction_t>& instructions,
+            const wf::render_target_t& target, wf::region_t& damage) override
+        {
+            for (auto& inst : children)
+            {
+                inst->schedule_instructions(instructions, target, damage);
+            }
+        }
+
+        void presentation_feedback(wf::output_t *output) override
+        {
+            for (auto& instance : children)
+            {
+                instance->presentation_feedback(output);
+            }
+        }
+
+        void compute_visibility(wf::output_t *output, wf::region_t& visible) override
+        {
+            for (auto& instance : children)
+            {
+                const int BIG_NUMBER    = 1e5;
+                wf::region_t big_region =
+                    wf::geometry_t{-BIG_NUMBER, -BIG_NUMBER, 2 * BIG_NUMBER, 2 * BIG_NUMBER};
+                instance->compute_visibility(output, big_region);
+            }
         }
     };
 };
@@ -366,31 +415,58 @@ struct drag_options_t
  *
  * Intended for use via wf::shared_data::ref_ptr_t.
  */
-class core_drag_t : public signal_provider_t
+class core_drag_t : public signal::provider_t
 {
     /**
      * Rebuild the wobbly model after a change in the scaling, so that the wobbly
      * model does not try to animate the scaling change itself.
      */
-    void rebuild_wobbly(wayfire_view view, wf::point_t grab, wf::pointf_t relative)
+    void rebuild_wobbly(wayfire_toplevel_view view, wf::point_t grab, wf::pointf_t relative)
     {
-        auto dim = wf::dimensions(view->get_bounding_box("wobbly"));
+        auto dim = wf::dimensions(wf::view_bounding_box_up_to(view, "wobbly"));
         modify_wobbly(view, find_geometry_around(dim, grab, relative));
     }
 
   public:
+    std::optional<wf::point_t> tentative_grab_position;
+
     /**
-     * Start drag.
+     * A button has been pressed which might start a drag action.
+     */
+    template<class Point>
+    void set_pending_drag(const Point& current_position)
+    {
+        this->tentative_grab_position = {(int)current_position.x, (int)current_position.y};
+    }
+
+    /**
+     * Check whether a motion event makes a sufficient drag so that the drag operation may start at all.
+     *
+     * Note that in some cases this functionality is not used at all, if the action for example was triggered
+     * by a binding.
+     */
+    bool should_start_pending_drag(wf::point_t current_position)
+    {
+        if (!tentative_grab_position.has_value())
+        {
+            return false;
+        }
+
+        return distance_to_grab_origin(current_position) > 5;
+    }
+
+    /**
+     * Start the actual dragging operation. Note: this should be called **after** set_pending_drag().
      *
      * @param grab_view The view which is being dragged.
      * @param grab_position The position of the input, in output-layout coordinates.
      * @param relative The position of the grab_position relative to view.
      */
-    void start_drag(wayfire_view grab_view, wf::point_t grab_position,
-        wf::pointf_t relative,
-        const drag_options_t& options)
+    void start_drag(wayfire_toplevel_view grab_view, wf::pointf_t relative, const drag_options_t& options)
     {
-        auto bbox = grab_view->get_bounding_box("wobbly");
+        wf::dassert(tentative_grab_position.has_value(),
+            "First, the drag operation should be set as pending!");
+        auto bbox = wf::view_bounding_box_up_to(grab_view, "wobbly");
         wf::point_t rel_grab_pos = {
             int(bbox.x + relative.x * bbox.width),
             int(bbox.y + relative.y * bbox.height),
@@ -403,6 +479,7 @@ class core_drag_t : public signal_provider_t
 
         this->view   = grab_view;
         this->params = options;
+        wf::get_core().default_wm->set_view_grabbed(view, true);
 
         auto target_views = get_target_views(grab_view, options.join_views);
         for (auto& v : target_views)
@@ -411,37 +488,34 @@ class core_drag_t : public signal_provider_t
             dragged.view = v;
 
             // Setup view transform
-            auto tr = std::make_unique<scale_around_grab_t>();
+
+            auto tr = std::make_shared<scale_around_grab_t>();
             dragged.transformer = {tr};
 
             tr->relative_grab = find_relative_grab(
-                v->get_bounding_box("wobbly"), rel_grab_pos);
-            tr->grab_position = grab_position;
+                wf::view_bounding_box_up_to(v, "wobbly"), rel_grab_pos);
+            tr->grab_position = *tentative_grab_position;
             tr->scale_factor.animate(options.initial_scale, options.initial_scale);
-
-            v->add_transformer(std::move(tr), move_drag_transformer);
+            v->get_transformed_node()->add_transformer(
+                tr, wf::TRANSFORMER_HIGHLEVEL - 1);
 
             // Hide the view, we will render it as an overlay
-            v->set_visible(false);
+            wf::scene::set_node_enabled(v->get_transformed_node(), false);
             v->damage();
 
             // Make sure that wobbly has the correct geometry from the start!
-            rebuild_wobbly(v, grab_position, dragged.transformer->relative_grab);
+            rebuild_wobbly(v, *tentative_grab_position, dragged.transformer->relative_grab);
 
             // TODO: make this configurable!
             start_wobbly_rel(v, dragged.transformer->relative_grab);
 
             this->all_views.push_back(dragged);
-            v->connect_signal("unmapped", &on_view_unmap);
+            v->connect(&on_view_unmap);
         }
 
         // Setup overlay hooks
-        for (auto& output : wf::get_core().output_layout->get_outputs())
-        {
-            output->store_data(
-                std::make_unique<output_data_t>(output, all_views));
-        }
-
+        render_node = std::make_shared<dragged_view_node_t>(all_views);
+        wf::scene::add_front(wf::get_core().scene(), render_node);
         wf::get_core().set_cursor("grabbing");
 
         // Set up snap-off
@@ -452,23 +526,23 @@ class core_drag_t : public signal_provider_t
                 set_tiled_wobbly(v.view, true);
             }
 
-            grab_origin = grab_position;
             view_held_in_place = true;
         }
     }
 
-    void start_drag(wayfire_view view, wf::point_t grab_position,
-        const drag_options_t& options)
+    void start_drag(wayfire_toplevel_view view, const drag_options_t& options)
     {
+        wf::dassert(tentative_grab_position.has_value(),
+            "First, the drag operation should be set as pending!");
+
         if (options.join_views)
         {
             view = get_toplevel(view);
         }
 
-        auto bbox = view->get_bounding_box() +
+        auto bbox = view->get_transformed_node()->get_bounding_box() +
             wf::origin(view->get_output()->get_layout_geometry());
-        start_drag(view, grab_position,
-            find_relative_grab(bbox, grab_position), options);
+        start_drag(view, find_relative_grab(bbox, *tentative_grab_position), options);
     }
 
     void handle_motion(wf::point_t to)
@@ -485,7 +559,7 @@ class core_drag_t : public signal_provider_t
 
                 snap_off_signal data;
                 data.focus_output = current_output;
-                emit_signal("snap-off", &data);
+                emit(&data);
             }
         }
 
@@ -497,7 +571,9 @@ class core_drag_t : public signal_provider_t
             move_wobbly(v.view, to.x, to.y);
             if (!view_held_in_place)
             {
+                v.view->get_transformed_node()->begin_transform_update();
                 v.transformer->grab_position = to;
+                v.view->get_transformed_node()->end_transform_update();
             }
         }
 
@@ -506,13 +582,18 @@ class core_drag_t : public signal_provider_t
 
     double distance_to_grab_origin(wf::point_t to) const
     {
-        auto offset = to - grab_origin;
-        const int dst_sq = offset.x * offset.x + offset.y * offset.y;
-        return std::sqrt(dst_sq);
+        return abs(to - *tentative_grab_position);
     }
 
     void handle_input_released()
     {
+        if (!view || all_views.empty())
+        {
+            this->tentative_grab_position = {};
+            // Input already released => don't do anything
+            return;
+        }
+
         // Store data for the drag done signal
         drag_done_signal data;
         data.grab_position = all_views.front().transformer->grab_position;
@@ -527,11 +608,8 @@ class core_drag_t : public signal_provider_t
         data.join_views     = params.join_views;
 
         // Remove overlay hooks and damage outputs BEFORE popping the transformer
-        for (auto& output : wf::get_core().output_layout->get_outputs())
-        {
-            output->get_data<output_data_t>()->apply_damage();
-            output->erase_data<output_data_t>();
-        }
+        wf::scene::remove_child(render_node);
+        render_node = nullptr;
 
         for (auto& v : all_views)
         {
@@ -539,8 +617,8 @@ class core_drag_t : public signal_provider_t
             auto rel_pos = v.transformer->relative_grab;
 
             // Restore view to where it was before
-            v.view->set_visible(true);
-            v.view->pop_transformer(move_drag_transformer);
+            wf::scene::set_node_enabled(v.view->get_transformed_node(), true);
+            v.view->get_transformed_node()->rem_transformer<scale_around_grab_t>();
 
             // Reset wobbly and leave it in output-LOCAL coordinates
             end_wobbly(v.view);
@@ -557,15 +635,18 @@ class core_drag_t : public signal_provider_t
         }
 
         // Reset our state
+        wf::get_core().default_wm->set_view_grabbed(view, false);
         view = nullptr;
         all_views.clear();
-        current_output     = nullptr;
-        view_held_in_place = false;
+        current_output = nullptr;
         wf::get_core().set_cursor("default");
 
         // Lastly, let the plugins handle what happens on drag end.
-        emit_signal("done", &data);
+        emit(&data);
+        view_held_in_place = false;
         on_view_unmap.disconnect();
+
+        this->tentative_grab_position = {};
     }
 
     void set_scale(double new_scale)
@@ -582,7 +663,7 @@ class core_drag_t : public signal_provider_t
     }
 
     // View currently being moved.
-    wayfire_view view;
+    wayfire_toplevel_view view;
 
     // Output where the action is happening.
     wf::output_t *current_output = NULL;
@@ -594,31 +675,50 @@ class core_drag_t : public signal_provider_t
     // Current parameters
     drag_options_t params;
 
-    // Grab origin, used for snap-off
-    wf::point_t grab_origin;
-
     // View is held in place, waiting for snap-off
     bool view_held_in_place = false;
 
+    std::shared_ptr<dragged_view_node_t> render_node;
+
     void update_current_output(wf::point_t grab)
     {
         wf::pointf_t origin = {1.0 * grab.x, 1.0 * grab.y};
-        auto output =
-            wf::get_core().output_layout->get_output_coords_at(origin, origin);
+        auto output = wf::get_core().output_layout->get_output_coords_at(origin, origin);
 
         if (output != current_output)
         {
+            if (current_output)
+            {
+                current_output->render->rem_effect(&on_pre_frame);
+            }
+
             drag_focus_output_signal data;
             data.previous_focus_output = current_output;
 
             current_output    = output;
             data.focus_output = output;
-            wf::get_core().focus_output(output);
-            emit_signal("focus-output", &data);
+            wf::get_core().seat->focus_output(output);
+            emit(&data);
+
+            if (output)
+            {
+                current_output->render->add_effect(&on_pre_frame, OUTPUT_EFFECT_PRE);
+            }
         }
     }
 
-    wf::signal_connection_t on_view_unmap = [=] (auto *ev)
+    wf::effect_hook_t on_pre_frame = [=] ()
+    {
+        for (auto& v : this->all_views)
+        {
+            if (v.transformer->scale_factor.running())
+            {
+                v.view->damage();
+            }
+        }
+    };
+
+    wf::signal::connection_t<view_unmapped_signal> on_view_unmap = [=] (auto *ev)
     {
         handle_input_released();
     };
@@ -642,7 +742,7 @@ inline void adjust_view_on_output(drag_done_signal *ev)
 
     if (parent->get_output() != ev->focused_output)
     {
-        wf::get_core().move_view_to_output(parent, ev->focused_output, false);
+        move_view_to_output(parent, ev->focused_output, false);
     }
 
     // Calculate the position we're leaving the view on
@@ -650,14 +750,14 @@ inline void adjust_view_on_output(drag_done_signal *ev)
     auto grab = ev->grab_position + output_delta;
 
     auto output_geometry = ev->focused_output->get_relative_geometry();
-    auto current_ws = ev->focused_output->workspace->get_current_workspace();
+    auto current_ws = ev->focused_output->wset()->get_current_workspace();
     wf::point_t target_ws{
         (int)std::floor(1.0 * grab.x / output_geometry.width),
         (int)std::floor(1.0 * grab.y / output_geometry.height),
     };
     target_ws = target_ws + current_ws;
 
-    auto gsize = ev->focused_output->workspace->get_workspace_grid_size();
+    auto gsize = ev->focused_output->wset()->get_workspace_grid_size();
     target_ws.x = wf::clamp(target_ws.x, 0, gsize.width - 1);
     target_ws.y = wf::clamp(target_ws.y, 0, gsize.height - 1);
 
@@ -672,8 +772,8 @@ inline void adjust_view_on_output(drag_done_signal *ev)
             continue;
         }
 
-        auto bbox = v.view->get_bounding_box("wobbly");
-        auto wm   = v.view->get_wm_geometry();
+        auto bbox = wf::view_bounding_box_up_to(v.view, "wobbly");
+        auto wm   = v.view->get_geometry();
 
         wf::point_t wm_offset = wf::origin(wm) + -wf::origin(bbox);
         bbox = wf::move_drag::find_geometry_around(
@@ -681,16 +781,16 @@ inline void adjust_view_on_output(drag_done_signal *ev)
 
         wf::point_t target = wf::origin(bbox) + wm_offset;
         v.view->move(target.x, target.y);
-        if (v.view->fullscreen)
+        if (v.view->pending_fullscreen())
         {
-            v.view->fullscreen_request(ev->focused_output, true, target_ws);
-        } else if (v.view->tiled_edges)
+            wf::get_core().default_wm->fullscreen_request(v.view, ev->focused_output, true, target_ws);
+        } else if (v.view->pending_tiled_edges())
         {
-            v.view->tile_request(v.view->tiled_edges, target_ws);
+            wf::get_core().default_wm->tile_request(v.view, v.view->pending_tiled_edges(), target_ws);
         }
 
         // check focus timestamp and select the last focused view to (re)focus
-        if (v.view->last_focus_timestamp > focus_view->last_focus_timestamp)
+        if (get_focus_timestamp(v.view) > get_focus_timestamp(focus_view))
         {
             focus_view = v.view;
         }
@@ -699,20 +799,20 @@ inline void adjust_view_on_output(drag_done_signal *ev)
     // Ensure that every view is visible on parent's main workspace
     for (auto& v : parent->enumerate_views())
     {
-        ev->focused_output->workspace->move_to_workspace(v, target_ws);
+        ev->focused_output->wset()->move_to_workspace(v, target_ws);
     }
 
-    ev->focused_output->focus_view(focus_view, true);
+    wf::get_core().default_wm->focus_raise_view(focus_view);
 }
 
 /**
  * Adjust the view's state after snap-off.
  */
-inline void adjust_view_on_snap_off(wayfire_view view)
+inline void adjust_view_on_snap_off(wayfire_toplevel_view view)
 {
-    if (view->tiled_edges && !view->fullscreen)
+    if (view->pending_tiled_edges() && !view->pending_fullscreen())
     {
-        view->tile_request(0);
+        wf::get_core().default_wm->tile_request(view, 0);
     }
 }
 }
