@@ -23,37 +23,50 @@ namespace wf
  * Core signals
  * -------------------------------------------------------------------------- */
 /**
- * name: startup-finished
  * on: core
- * when: Emitted when the Wayfire initialization has been completed and the main
- *   loop is about to start.
- * argument: unused
+ * when: Emitted when the wlroots backend has been started.
  */
+struct core_backend_started_signal
+{};
+
+/**
+ * on: core
+ * when: Emitted when the Wayfire initialization has been completed and the main loop is about to start.
+ */
+struct core_startup_finished_signal
+{};
 
 /**
- * name: shutdown
  * on: core
  * when: Right before the shutdown sequence starts.
- * argument: unused
  */
+struct core_shutdown_signal
+{};
 
 class input_device_t;
 /**
- * name: input-device-added, input-device-removed
  * on: core
- * when: Whenever a new input device is added or removed.
+ * when: Whenever a new input device is added.
  */
-struct input_device_signal : public wf::signal_data_t
+struct input_device_added_signal
+{
+    nonstd::observer_ptr<input_device_t> device;
+};
+
+/**
+ * on: core
+ * when: Whenever an input device is removed.
+ */
+struct input_device_removed_signal
 {
     nonstd::observer_ptr<input_device_t> device;
 };
 
 /**
- * name: tablet-mode, lid-state
  * on: core
  * when: When the corresponding switch device state changes.
  */
-struct switch_signal : public wf::signal_data_t
+struct switch_signal
 {
     /** The switch device */
     nonstd::observer_ptr<input_device_t> device;
@@ -61,6 +74,7 @@ struct switch_signal : public wf::signal_data_t
     bool state;
 };
 
+
 /**
  * Describes the various ways in which core should handle an input event.
  */
@@ -76,13 +90,17 @@ enum class input_event_processing_mode_t
      * the event to the client.
      */
     NO_CLIENT,
+    /**
+     * Core should not process this event at all.
+     */
+    IGNORE,
 };
 
 /**
- * name:
+ * Emitted for the following events:
  *   pointer_motion, pointer_motion_absolute, pointer_button, pointer_axis,
- *   pointer_swipe_begin, pointer_swipe_update, pointer_swipe_end,
- *   pointer_pinch_begin, pointer_pinch_update, pointer_pinch_end,
+ *   pointer_swipe_begin, pointer_swipe_update, pointer_swipe_end, pointer_pinch_begin, pointer_pinch_update,
+ *   pointer_pinch_end, pointer_hold_begin, pointer_hold_end,
  *   keyboard_key,
  *   touch_down, touch_up, touch_motion,
  *   tablet_proximity, tablet_axis, tablet_button, tablet_tip
@@ -97,14 +115,11 @@ enum class input_event_processing_mode_t
  *   will be used instead. However plugins which modify the event must ensure
  *   that subsequent events are adjusted accordingly as well.
  *
- *   The same signals are emitted with a _post suffix after the event handling
- *   by core has finished.
- *
  * example: The pointer_motion event is emitted with data of type
- *   input_event_signal<wlr_event_pointer_motion>
+ *   input_event_signal<wlr_pointer_motion_event>
  */
 template<class wlr_event_t>
-struct input_event_signal : public wf::signal_data_t
+struct input_event_signal
 {
     /* The event as it has arrived from wlroots */
     wlr_event_t *event;
@@ -117,46 +132,74 @@ struct input_event_signal : public wf::signal_data_t
      * pointer_button, keyboard_key, touch_down
      */
     input_event_processing_mode_t mode = input_event_processing_mode_t::FULL;
+
+    /**
+     * The wlr device which triggered the event. May be NULL for events which are merged together for ex. via
+     * wlr-cursor.
+     */
+    wlr_input_device *device;
 };
 
 /**
- * name: drag-started, drag-stopped
- * on: core
- * when: When a DnD action is started/stopped
+ * Same as @input_event_signal, but emitted after bindings have been handled and before the event is sent
+ * to the client (if at all).
+ *
+ * Note: currently only keyboard_key events support this signal.
+ * TODO: add the event for the rest of the event types.
  */
-struct dnd_signal : public wf::signal_data_t
+template<class wlr_event_t>
+struct pre_client_input_event_signal
 {
-    /** The DnD icon */
-    wf::surface_interface_t *icon;
+    wlr_event_t *event;
+    wlr_input_device *device;
+
+    /** Last opportunity for plugins to influence the processing of this event. */
+    bool carried_out = false;
+
+    /** The node which will receive the event. May be NULL. */
+    wf::scene::node_ptr focus_node;
 };
 
 /**
- * name: surface-mapped, surface-unmapped
- * on: core
- * when: Whenever a surface map state changes. This must be emitted for all
- *   surfaces regardless of their type (view, subsurface, etc).
+ * Same as @input_event_signal, but emitted after the event has been handled.
  */
-struct surface_map_state_changed_signal : public wf::signal_data_t
+template<class wlr_event_t>
+struct post_input_event_signal
 {
-    wf::surface_interface_t *surface;
+    wlr_event_t *event;
+    wlr_input_device *device;
 };
 
 /**
- * name: reload-config
  * on: core
  * when: When the config file is reloaded
- * argument: unused
  */
+struct reload_config_signal
+{};
 
 /**
- * name: keyboard-focus-changed
  * on: core
  * when: Keyboard focus is changed (may change to nullptr).
  */
-struct keyboard_focus_changed_signal : public wf::signal_data_t
+struct keyboard_focus_changed_signal
 {
-    wayfire_view view;
-    wlr_surface *surface;
+    wf::scene::node_ptr new_focus;
+};
+
+/**
+ * on: core
+ * when: Seat activity has happened after being idle.
+ */
+struct seat_activity_signal
+{};
+
+/**
+ * on: core
+ * when: idle inhibit changed.
+ */
+struct idle_inhibit_changed_signal
+{
+    bool inhibit;
 };
 
 /* ----------------------------------------------------------------------------/
@@ -164,34 +207,33 @@ struct keyboard_focus_changed_signal : public wf::signal_data_t
  * -------------------------------------------------------------------------- */
 
 /** Base class for all output signals. */
-struct _output_signal : public wf::signal_data_t
+
+/**
+ * on: output-layout
+ * when: Each time a new output is added.
+ */
+struct output_added_signal
 {
     wf::output_t *output;
 };
 
-/** @return The output in the signal. Must be an _output_signal. */
-output_t *get_signaled_output(signal_data_t *data);
-
 /**
- * name: output-added
- * on: output-layout
- * when: Each time a new output is added.
- */
-using output_added_signal = _output_signal;
-
-/**
- * name: pre-remove
  * on: output, output-layout(output-)
  * when: Emitted just before starting the destruction procedure for an output.
  */
-using output_pre_remove_signal = _output_signal;
+struct output_pre_remove_signal
+{
+    wf::output_t *output;
+};
 
 /**
- * name: output-removed
  * on: output-layout
  * when: Each time a new output is added.
  */
-using output_removed_signal = _output_signal;
+struct output_removed_signal
+{
+    wf::output_t *output;
+};
 
 enum output_config_field_t
 {
@@ -210,16 +252,22 @@ enum output_config_field_t
 struct output_state_t;
 
 /**
- * name: configuration-changed
- * on: output, output-layout(output-)
- * when: Each time the output's source, mode, scale, transform and/or position
- *   changes.
+ * on: output-layout
+ * when: Each time the configuration of the output layout changes.
  */
-struct output_configuration_changed_signal : public _output_signal
+struct output_layout_configuration_changed_signal
+{};
+
+/**
+ * on: output
+ * when: Each time the output's source, mode, scale, transform and/or position changes.
+ */
+struct output_configuration_changed_signal
 {
-    output_configuration_changed_signal(const wf::output_state_t& st) :
-        state(st)
+    wf::output_t *output;
+    output_configuration_changed_signal(const wf::output_state_t& st) : state(st)
     {}
+
     /**
      * Which output attributes actually changed.
      * A bitwise OR of output_config_field_t.
@@ -233,39 +281,64 @@ struct output_configuration_changed_signal : public _output_signal
 };
 
 /**
- * name: gain-focus
  * on: output, core(output-)
  * when: Immediately after the output becomes focused.
  */
-using output_gain_focus_signal = _output_signal;
+struct output_gain_focus_signal
+{
+    wf::output_t *output;
+};
+
+/**
+ * on: output, core
+ * when: When an output activates or deactivates. Note that not all plugin actions are reflected with this
+ *   signal. A plugin activates on an output usually if it (temporarily) changes the way Wayfire works like
+ *   Expo, Scale, Vswitch. One-shot actions like command or wsets do not send this signal.
+ */
+struct output_plugin_activated_changed_signal
+{
+    // The output on which the plugin was activated. May be NULL if the plugin works globally.
+    wf::output_t *output;
+    // The name of the plugin.
+    std::string plugin_name;
+    // Whether the plugin was activated (true) or deactivated (false).
+    bool activated;
+};
 
 /* ----------------------------------------------------------------------------/
  * Output rendering signals (see also wayfire/workspace-stream.hpp)
  * -------------------------------------------------------------------------- */
 /**
- * name: start-rendering
  * on: output
  * when: Whenever the output is ready to start rendering. This can happen
  *   either on output creation or whenever all inhibits in wayfire-shell have
  *   been removed.
  */
-using output_start_rendering_signal = _output_signal;
+struct output_start_rendering_signal
+{
+    wf::output_t *output;
+};
 
 /* ----------------------------------------------------------------------------/
  * Output workspace signals
  * -------------------------------------------------------------------------- */
 
 /**
- * name: workspace-changed
  * on: output
+ * when: Whenever the current workspace set of the output changes.
+ */
+struct workspace_set_changed_signal
+{
+    std::shared_ptr<workspace_set_t> new_wset;
+    wf::output_t *output;
+};
+
+/**
+ * on: workspace set, output
  * when: Whenever the current workspace on the output has changed.
  */
-struct workspace_changed_signal : public wf::signal_data_t
+struct workspace_changed_signal
 {
-    /** For workspace-change-request, whether the request has already been
-     * handled. */
-    bool carried_out;
-
     /** Previously focused workspace */
     wf::point_t old_viewport;
 
@@ -277,14 +350,25 @@ struct workspace_changed_signal : public wf::signal_data_t
 };
 
 /**
- * name: workspace-change-request
  * on: output
  * when: Whenever a workspace change is requested by core or by a plugin.
  *   This can be used by plugins who wish to handle workspace changing
  *   themselves, for ex. if animating the transition.
  */
-struct workspace_change_request_signal : public workspace_changed_signal
+struct workspace_change_request_signal
 {
+    /** Previously focused workspace */
+    wf::point_t old_viewport;
+
+    /** Workspace that is to be focused or became focused */
+    wf::point_t new_viewport;
+
+    /** The output this is happening on */
+    wf::output_t *output;
+
+    /** Whether the request has already been handled. */
+    bool carried_out;
+
     /**
      * A list of views whose geometry should remain stationary.
      * The caller is responsible for ensuring that this doesn't move the views
@@ -293,138 +377,120 @@ struct workspace_change_request_signal : public workspace_changed_signal
      * Note that the views might still be moved if a previous workspace change
      * request is being serviced.
      */
-    std::vector<wayfire_view> fixed_views;
+    std::vector<wayfire_toplevel_view> fixed_views;
+};
+
+/**
+ * on: workspace set
+ * when: Whenever the workspace grid size changes.
+ */
+struct workspace_grid_changed_signal
+{
+    /** The grid size before the change. */
+    wf::dimensions_t old_grid_size;
+
+    /** The grid size after the change. */
+    wf::dimensions_t new_grid_size;
 };
 
 /**
- * name: workarea-changed
  * on: output
  * when: Whenever the available workarea changes.
  */
-struct workarea_changed_signal : public wf::signal_data_t
+struct workarea_changed_signal
 {
+    wf::output_t *output;
     wf::geometry_t old_workarea;
     wf::geometry_t new_workarea;
 };
 
 /**
- * name: fullscreen-layer-focused
  * on: output
  * when: Whenever a fullscreen view is promoted on top of the other layers.
- * argument: The event data pointer is null if there are no promoted views,
- *   and not-null otherwise.
  */
-
-/**
- * name: stack-order-changed
- * on: output, core(output-)
- * when: Whenever the stacking order of views on the output changes. Note that
- *   the signal can be a false positive, i.e the stacking order did not really
- *   change. However, there will be no cases where the order changed but the
- *   signal was not fired.
- */
-using stack_order_changed_signal = _output_signal;
-
-/* ----------------------------------------------------------------------------/
- * Surface signals
- * -------------------------------------------------------------------------- */
-struct _subsurface_signal : public wf::signal_data_t
+struct fullscreen_layer_focused_signal
 {
-    nonstd::observer_ptr<surface_interface_t> main_surface;
-    nonstd::observer_ptr<surface_interface_t> subsurface;
+    bool has_promoted;
 };
 
-/**
- * name: subsurface-added
- * on: surface
- * when: Emitted when a subsurface is added to the given surface.
- */
-using subsurface_added_signal = _subsurface_signal;
-
-/**
- * name: subsurface-removed
- * on: surface
- * when: Emitted immediately before removing a subsurface from the surface.
- */
-using subsurface_removed_signal = _subsurface_signal;
-
 /* ----------------------------------------------------------------------------/
  * View signals
  * -------------------------------------------------------------------------- */
-
-/** Base class for all view signals. */
-struct _view_signal : public wf::signal_data_t
+/**
+ * on: view, output, core
+ * when: After the view becomes mapped. This signal must also be emitted from all compositor views.
+ */
+struct view_mapped_signal
 {
     wayfire_view view;
-};
 
-/**
- * @return The view contained in the signal data, it must be
- * a _view_signal.
- */
-wayfire_view get_signaled_view(wf::signal_data_t *data);
-
-/**
- * name: mapped
- * on: view, output(view-)
- * when: After the view becomes mapped. This signal must also be emitted from
- *   all compositor views.
- */
-struct view_mapped_signal : public _view_signal
-{
     /* Indicates whether the position already has its initial position */
     bool is_positioned = false;
 };
 
 /**
- * name: pre-unmapped
- * on: view, output(view-)
+ * on: view, output, core
  * when: Immediately before unmapping a mapped view. The signal may not be
  *   emitted from all views, but it is necessary for unmap animations to work.
  */
-using view_pre_unmap_signal = _view_signal;
+struct view_pre_unmap_signal
+{
+    wayfire_view view;
+};
 
 /**
  * name: unmapped
- * on: view, output(view-)
+ * on: view, output, core
  * when: After a previously mapped view becomes unmapped. This must be emitted
  *   for all views.
  */
-using view_unmapped_signal = _view_signal;
+struct view_unmapped_signal
+{
+    wayfire_view view;
+};
 
 /**
- * name: set-output
- * on: view
- * when: Immediately after the view's output changes. Note that child views may
- *   still be on the old output.
- * argument: The old output of the view.
+ * on: view, new output, core
+ * when: Immediately after the view's output changes. Note that child views may still be on the old output.
  */
-using view_set_output_signal = _output_signal;
+struct view_set_output_signal
+{
+    wayfire_view view;
+    // The previous output of the view.
+    wf::output_t *output;
+};
 
 /* ----------------------------------------------------------------------------/
  * View state signals
  * -------------------------------------------------------------------------- */
-
 /**
- * name: minimized
- * on: view, output(view-)
- * when: After the view's minimized state changes.
+ * on: view
+ * when: After the view's parent changes.
  */
-struct view_minimized_signal : public _view_signal
+struct view_parent_changed_signal
 {
-    /** true is minimized, false is restored */
-    bool state;
+    wayfire_toplevel_view view;
+};
+
+/**
+ * on: view, output(view-)
+ * when: After the view's minimized state changes.
+ */
+struct view_minimized_signal
+{
+    wayfire_toplevel_view view;
 };
 
 /**
- * name: view-minimize-request
  * on: output
  * when: Emitted whenever some entity requests that the view's minimized state
  *   changes. If no plugin is available to service the request, it is carried
  *   out by core. See view_interface_t::minimize_request()
  */
-struct view_minimize_request_signal : public _view_signal
+struct view_minimize_request_signal
 {
+    wayfire_toplevel_view view;
+
     /** true is minimized, false is restored */
     bool state;
 
@@ -436,12 +502,22 @@ struct view_minimize_request_signal : public _view_signal
 };
 
 /**
- * name: tiled
+ * on: view
+ * when: After the view's activated state changes.
+ */
+struct view_activated_state_signal
+{
+    wayfire_toplevel_view view;
+};
+
+/**
  * on: view, output(view-)
  * when: After the view's tiled edges change.
  */
-struct view_tiled_signal : public _view_signal
+struct view_tiled_signal
 {
+    wayfire_toplevel_view view;
+
     /** Previously tiled edges */
     uint32_t old_edges;
     /** Currently tiled edges */
@@ -449,14 +525,15 @@ struct view_tiled_signal : public _view_signal
 };
 
 /**
- * name: view-tile-request
  * on: output
  * when: Emitted whenever some entity requests that the view's tiled edges
  *   change. If no plugin is available to service the request, it is carried
  *   out by core. See view_interface_t::tile_request()
  */
-struct view_tile_request_signal : public _view_signal
+struct view_tile_request_signal
 {
+    wayfire_toplevel_view view;
+
     /** The desired edges */
     uint32_t edges;
 
@@ -479,28 +556,36 @@ struct view_tile_request_signal : public _view_signal
     bool carried_out = false;
 };
 
-
 /**
- * name: fullscreen
  * on: view, output(view-)
  * when: After the view's fullscreen state changes.
  */
-struct view_fullscreen_signal : public _view_signal
+struct view_fullscreen_signal
 {
+    wayfire_toplevel_view view;
+    bool state;
+};
+
+/**
+ * on: output
+ * when: Emitted whenever some entity requests that the view's fullscreen state
+ *   change. If no plugin is available to service the request, it is carried
+ *   out by core. See view_interface_t::fullscreen_request()
+ */
+struct view_fullscreen_request_signal
+{
+    wayfire_toplevel_view view;
+
     /** The desired fullscreen state */
     bool state;
 
     /**
-     * For view-fullscreen-request:
-     *
      * Whether some plugin will service the fullscreen request, in which case
      * other plugins and core should ignore the request.
      */
     bool carried_out = false;
 
     /**
-     * For view-fullscreen-request:
-     *
      * The geometry the view should have. This is for example the last geometry
      * a view had before being fullscreened. The given geometry is only a hint
      * by core and plugins may override it. It may also be undefined (0,0 0x0).
@@ -508,29 +593,20 @@ struct view_fullscreen_signal : public _view_signal
     wf::geometry_t desired_size;
 
     /**
-     * For view-fullscreen-request:
-     *
      * The target workspace of the operation.
      */
     wf::point_t workspace;
 };
 
-/**
- * name: view-fullscreen-request
- * on: output
- * when: Emitted whenever some entity requests that the view's fullscreen state
- *   change. If no plugin is available to service the request, it is carried
- *   out by core. See view_interface_t::fullscreen_request()
- */
-using view_fullscreen_request_signal = view_fullscreen_signal;
 
 /**
- * name: view-focus-request
  * on: view, core
  * when: Emitted whenever some entity (typically a panel) wants to focus the view.
  */
-struct view_focus_request_signal : public _view_signal
+struct view_focus_request_signal
 {
+    wayfire_view view;
+
     /** Set to true if core and other plugins should not handle this request. */
     bool carried_out = false;
 
@@ -543,193 +619,178 @@ struct view_focus_request_signal : public _view_signal
  * on: view, output(view-)
  * when: Whenever the view's sticky state changes.
  */
-using view_set_sticky_signal = _view_signal;
+struct view_set_sticky_signal
+{
+    wayfire_toplevel_view view;
+};
 
 /**
- * name: title-changed
- * on: view
+ * on: view, core
  * when: After the view's title has changed.
  */
-using title_changed_signal = _view_signal;
+struct view_title_changed_signal
+{
+    wayfire_view view;
+};
 
 /**
- * name: app-id-changed
- * on: view
+ * on: view, core
  * when: After the view's app-id has changed.
  */
-using app_id_changed_signal = _view_signal;
+struct view_app_id_changed_signal
+{
+    wayfire_view view;
+};
 
 /**
- * name: view-show-window-menu
  * on: output, core
  * when: To show a menu with window related actions.
  */
-struct view_show_window_menu_signal : public _view_signal
+struct view_show_window_menu_signal
 {
+    wayfire_view view;
+
     /** The position as requested by the client, in surface coordinates */
     wf::point_t relative_position;
 };
 
 /**
- * name: geometry-changed
  * on: view, output(view-), core(view-)
  * when: Whenever the view's wm geometry changes.
  */
-struct view_geometry_changed_signal : public _view_signal
+struct view_geometry_changed_signal
 {
+    wayfire_toplevel_view view;
+
     /** The old wm geometry */
     wf::geometry_t old_geometry;
 };
 
 /**
- * name: region-damaged
- * on: view
- * when: Whenever a region of the view becomes damaged, for ex. when the client
- *   updates its contents.
- * argument: Unused.
+ * on: output
+ * when: Whenever the view's workspace changes. (Every plugin changing the
+ *   view's workspace should emit this signal).
  */
+struct view_change_workspace_signal
+{
+    wayfire_toplevel_view view;
+
+    wf::point_t from, to;
+
+    /**
+     * Indicates whether the old workspace is known.
+     * If false, then the `from` field should be ignored.
+     */
+    bool old_workspace_valid = true;
+};
 
 /**
- * name: decoration-state-updated
- * on: view, output(view-)
+ * on: view, core
  * when: Whenever the value of view::should_be_decorated() changes.
  */
-using view_decoration_state_updated_signal = _view_signal;
+struct view_decoration_state_updated_signal
+{
+    wayfire_toplevel_view view;
+};
 
 /**
- * name: decoration-changed
- * on: view
- * when: Whenever the view's decoration changes.
- * argument: unused.
- */
-
-/**
- * name: ping-timeout
  * on: view
  * when: Whenever the client fails to respond to a ping request within
  *   the expected time(10 seconds).
  */
-using view_ping_timeout_signal = _view_signal;
+struct view_ping_timeout_signal
+{
+    wayfire_view view;
+};
 
 /* ----------------------------------------------------------------------------/
  * View <-> output signals
  * -------------------------------------------------------------------------- */
 
 /**
- * name: view-attached
- * on: output
- * when: As soon as the view's output is set to the given output. This is the
- *   first point where the view is considered to be part of that output.
- */
-using view_attached_signal = _view_signal;
-
-/**
- * name: view-layer-attached
- * on: output
- * when: Emitted when the view is added to a layer in the output's workspace
- *   manager and it was in no layer previously.
- */
-using view_layer_attached_signal = _view_signal;
-
-/**
- * name: view-detached
- * on: output
- * when: Emitted when the view's output is about to be changed to another one.
- *   This is the last point where the view is considered to be part of the given
- *   output.
- */
-using view_detached_signal = _view_signal;
-
-/**
- * name: view-layer-detached
- * on: output
- * when: Emitted when the view is removed from a layer but is not added to
- *   another.
- */
-using view_layer_detached_signal = _view_signal;
-
-/**
- * name: view-pre-moved-to-output
  * on: core
- * when: Immediately before the view is moved to another output. The usual
- *   sequence when moving views to another output is:
- *
- *   pre-moved-to-output -> layer-detach -> detach ->
- *      attach -> layer-attach -> moved-to-output
+ * when: Immediately before the view is moved to a different workspace set.
+ *   view-moved-to-set is emitted afterwards.
  */
-struct view_pre_moved_to_output_signal : public signal_data_t
+struct view_pre_moved_to_wset_signal
 {
     /* The view being moved */
-    wayfire_view view;
-    /* The output the view was on, may be NULL. */
-    wf::output_t *old_output;
-    /* The output the view is being moved to. */
-    wf::output_t *new_output;
+    wayfire_toplevel_view view;
+    /* The old wset the view was on, may be NULL. */
+    std::shared_ptr<wf::workspace_set_t> old_wset;
+    /* The new wset the view is being moved to, may be NULL. */
+    std::shared_ptr<wf::workspace_set_t> new_wset;
 };
 
 /**
- * name: view-moved-to-output
  * on: core
- * when: After the view has been moved to a new output.
+ * when: After the view has been moved to a new wset.
  */
-using view_moved_to_output_signal = view_pre_moved_to_output_signal;
+struct view_moved_to_wset_signal
+{
+    /* The view being moved */
+    wayfire_toplevel_view view;
+    /* The old wset the view was on, may be NULL. */
+    std::shared_ptr<wf::workspace_set_t> old_wset;
+    /* The new wset the view is being moved to, may be NULL. */
+    std::shared_ptr<wf::workspace_set_t> new_wset;
+};
 
 /**
- * name: view-disappeared
  * on: output
- * when: This is a signal which combines view-unmapped, view-detached and
- *   view-minimized, and is emitted together with each of these three. Semantic
- *   meaning is that the view is no longer available for focus, interaction with
- *   the user, etc.
+ * when: This signal is a combination of the unmapped, minimized and set-output signals. In the latter case,
+ *   the signal is emitted on the view's previous output. The meaning of this signal is that the view is no
+ *   longer available for focus, interaction with the user, etc. on the output where it used to be.
  */
-using view_disappeared_signal = _view_signal;
+struct view_disappeared_signal
+{
+    wayfire_view view;
+};
 
 /**
- * name: view-focused
- * on: output
- * when: As soon as the output focus changes.
- * argument: The newly focused view.
- */
-using focus_view_signal = _view_signal;
-
-/**
- * name: view-move-request
  * on: output
  * when: Whenever an interactive move is requested on the view. See also
  *   view_interface_t::move_request()
  */
-using view_move_request_signal = _view_signal;
+struct view_move_request_signal
+{
+    wayfire_toplevel_view view;
+};
 
 /**
- * name: view-resize-request
  * on: output
  * when: Whenever an interactive resize is requested on the view. See also
  *   view_interface_t::resize_request()
  */
-struct view_resize_request_signal : public _view_signal
+struct view_resize_request_signal
 {
+    wayfire_toplevel_view view;
+
     /** The requested resize edges */
     uint32_t edges;
 };
 
 /**
- * name: hints-changed
  * on: view and core(view-)
  * when: the client indicates the views hints have changed (example urgency hint).
  */
-struct view_hints_changed_signal : public _view_signal
+struct view_hints_changed_signal
 {
+    wayfire_view view;
+
     bool demands_attention = false;
 };
 
 /**
- * name: view-system-bell
  * on: core
  * when: Whenever a client wants to invoke the system bell if such is available.
  *   Note the system bell may or may not be tied to a particular view, so the
  *   signal may be emitted with a nullptr view.
  */
-using view_system_bell_signal = _view_signal;
+struct view_system_bell_signal
+{
+    wayfire_view view;
+};
 }
 
 #endif
