@@ -1,14 +1,17 @@
 #include <wayfire/util/log.hpp>
 #include <map>
 #include "opengl-priv.hpp"
+#include "wayfire/geometry.hpp"
 #include "wayfire/output.hpp"
 #include "core-impl.hpp"
 #include "config.h"
 #include <wayfire/nonstd/wlroots-full.hpp>
+#include <set>
 
 #include <glm/gtc/matrix_transform.hpp>
 
 #include "shaders.tpp"
+#include "wayfire/region.hpp"
 
 const char *gl_error_string(const GLenum err)
 {
@@ -40,7 +43,7 @@ void gl_call(const char *func, uint32_t line, const char *glfunc)
     }
 
     LOGE("gles2: function ", glfunc, " in ", func, " line ", line, ": ",
-        gl_error_string(glGetError()));
+        gl_error_string(err));
 }
 
 namespace OpenGL
@@ -85,17 +88,31 @@ GLuint compile_program(std::string vertex_source, std::string frag_source)
     GL_CALL(glAttachShader(result_program, fragment_shader));
     GL_CALL(glLinkProgram(result_program));
 
+    int s = GL_FALSE;
+#define LENGTH 1024 * 128
+    char log[LENGTH];
+    GL_CALL(glGetProgramiv(result_program, GL_LINK_STATUS, &s));
+    GL_CALL(glGetProgramInfoLog(result_program, LENGTH, NULL, log));
+
+    if (s == GL_FALSE)
+    {
+        LOGE("Failed to link vertex shader:\n", vertex_source,
+            "\nFragment shader:\n", frag_source,
+            "\nLinker output:\n", log);
+
+        GL_CALL(glDeleteProgram(result_program));
+    }
+
     /* won't be really deleted until program is deleted as well */
     GL_CALL(glDeleteShader(vertex_shader));
     GL_CALL(glDeleteShader(fragment_shader));
-
-    return result_program;
+    return (s == GL_FALSE) ? 0 : result_program;
 }
 
 void init()
 {
     render_begin();
-    // enable_gl_synchronuous_debug()
+    // enable_gl_synchronous_debug()
     program.compile(default_vertex_shader_source,
         default_fragment_shader_source);
 
@@ -115,16 +132,19 @@ void fini()
 
 namespace
 {
-uint32_t current_output_fb = 0;
+wf::output_t *current_output = NULL;
+uint32_t current_output_fb   = 0;
 }
 
-void bind_output(uint32_t fb)
+void bind_output(wf::output_t *output, uint32_t fb)
 {
+    current_output    = output;
     current_output_fb = fb;
 }
 
-void unbind_output()
+void unbind_output(wf::output_t *output)
 {
+    current_output    = NULL;
     current_output_fb = 0;
 }
 
@@ -213,7 +233,7 @@ void render_transformed_texture(wf::texture_t texture,
 }
 
 void render_texture(wf::texture_t texture,
-    const wf::framebuffer_t& framebuffer,
+    const wf::render_target_t& framebuffer,
     const wf::geometry_t& geometry, glm::vec4 color, uint32_t bits)
 {
     render_transformed_texture(texture, geometry,
@@ -273,7 +293,7 @@ void render_begin()
     GL_CALL(glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA));
 }
 
-void render_begin(const wf::framebuffer_base_t& fb)
+void render_begin(const wf::framebuffer_t& fb)
 {
     render_begin();
     fb.bind();
@@ -322,7 +342,7 @@ static std::string framebuffer_status_to_str(
     }
 }
 
-bool wf::framebuffer_base_t::allocate(int width, int height)
+bool wf::framebuffer_t::allocate(int width, int height)
 {
     bool first_allocate = false;
     if (fb == (uint32_t)-1)
@@ -383,50 +403,20 @@ bool wf::framebuffer_base_t::allocate(int width, int height)
     return is_resize || first_allocate;
 }
 
-void wf::framebuffer_base_t::copy_state(wf::framebuffer_base_t&& other)
-{
-    this->viewport_width  = other.viewport_width;
-    this->viewport_height = other.viewport_height;
-
-    this->fb  = other.fb;
-    this->tex = other.tex;
-
-    other.reset();
-}
-
-wf::framebuffer_base_t::framebuffer_base_t(wf::framebuffer_base_t&& other)
-{
-    copy_state(std::move(other));
-}
-
-wf::framebuffer_base_t& wf::framebuffer_base_t::operator =(
-    wf::framebuffer_base_t&& other)
-{
-    if (this == &other)
-    {
-        return *this;
-    }
-
-    release();
-    copy_state(std::move(other));
-
-    return *this;
-}
-
-void wf::framebuffer_base_t::bind() const
+void wf::framebuffer_t::bind() const
 {
     GL_CALL(glBindFramebuffer(GL_DRAW_FRAMEBUFFER, fb));
     GL_CALL(glViewport(0, 0, viewport_width, viewport_height));
 }
 
-void wf::framebuffer_base_t::scissor(wlr_box box) const
+void wf::framebuffer_t::scissor(wlr_box box) const
 {
     GL_CALL(glEnable(GL_SCISSOR_TEST));
     GL_CALL(glScissor(box.x, viewport_height - box.y - box.height,
         box.width, box.height));
 }
 
-void wf::framebuffer_base_t::release()
+void wf::framebuffer_t::release()
 {
     if ((fb != uint32_t(-1)) && (fb != 0))
     {
@@ -441,14 +431,14 @@ void wf::framebuffer_base_t::release()
     reset();
 }
 
-void wf::framebuffer_base_t::reset()
+void wf::framebuffer_t::reset()
 {
     fb  = -1;
     tex = -1;
     viewport_width = viewport_height = 0;
 }
 
-wlr_box wf::framebuffer_t::framebuffer_box_from_geometry_box(wlr_box box) const
+wlr_box wf::render_target_t::framebuffer_box_from_geometry_box(wlr_box box) const
 {
     /* Step 1: Make relative to the framebuffer */
     box.x -= this->geometry.x;
@@ -458,15 +448,6 @@ wlr_box wf::framebuffer_t::framebuffer_box_from_geometry_box(wlr_box box) const
     wlr_box scaled = box * scale;
 
     /* Step 3: rotate */
-    if (has_nonstandard_transform)
-    {
-        // TODO: unimplemented, but also unused for now
-        LOGE("unimplemented reached: framebuffer_box_from_geometry_box"
-             " with has_nonstandard_transform");
-
-        return {0, 0, 0, 0};
-    }
-
     int width = viewport_width, height = viewport_height;
     if (wl_transform & 1)
     {
@@ -479,26 +460,78 @@ wlr_box wf::framebuffer_t::framebuffer_box_from_geometry_box(wlr_box box) const
 
     wlr_box_transform(&result, &scaled, transform, width, height);
 
+    if (subbuffer)
+    {
+        result = scale_box({0, 0, viewport_width, viewport_height},
+            subbuffer.value(), result);
+    }
+
     return result;
 }
 
-glm::mat4 wf::framebuffer_t::get_orthographic_projection() const
+wf::region_t wf::render_target_t::framebuffer_region_from_geometry_region(const wf::region_t& region) const
+{
+    wf::region_t result;
+    for (const auto& rect : region)
+    {
+        result |= framebuffer_box_from_geometry_box(wlr_box_from_pixman_box(rect));
+    }
+
+    return result;
+}
+
+glm::mat4 wf::render_target_t::get_orthographic_projection() const
 {
     auto ortho = glm::ortho(1.0f * geometry.x,
         1.0f * geometry.x + 1.0f * geometry.width,
         1.0f * geometry.y + 1.0f * geometry.height,
         1.0f * geometry.y);
 
-    return this->transform * ortho;
+    return gl_to_framebuffer() * ortho;
 }
 
-void wf::framebuffer_t::logic_scissor(wlr_box box) const
+glm::mat4 wf::render_target_t::gl_to_framebuffer() const
+{
+    if (subbuffer)
+    {
+        auto sub = subbuffer.value();
+
+        float scale_x = 1.0 * sub.width / viewport_width;
+        float scale_y = 1.0 * sub.height / viewport_height;
+
+        // Translation is calculated between the midpoint of the whole buffer
+        // and the midpoint of the subbuffer, then scaled to NDC.
+        float half_w = viewport_width / 2.0;
+        float half_h = viewport_height / 2.0;
+
+        float translate_x = ((sub.x + sub.width / 2.0) - half_w) / half_w;
+        float translate_y = (half_h - sub.y - sub.height / 2.0) / half_h;
+
+        glm::mat4 scale = glm::scale(glm::mat4(1.0),
+            glm::vec3(scale_x, scale_y, 1.0));
+        glm::mat4 translate = glm::translate(glm::mat4(1.0),
+            glm::vec3(translate_x, translate_y, 0.0));
+
+        return translate * scale * this->transform;
+    }
+
+    return this->transform;
+}
+
+void wf::render_target_t::logic_scissor(wlr_box box) const
 {
     scissor(framebuffer_box_from_geometry_box(box));
 }
 
+wf::render_target_t wf::render_target_t::translated(wf::point_t offset) const
+{
+    render_target_t copy = *this;
+    copy.geometry = copy.geometry + offset;
+    return copy;
+}
+
 /* look up the actual values of wl_output_transform enum
- * All _flipped transforms have values (regular_transfrom + 4) */
+ * All _flipped transforms have values (regular_transform + 4) */
 glm::mat4 get_output_matrix_from_transform(wl_output_transform transform)
 {
     glm::mat4 scale = glm::mat4(1.0);
@@ -542,7 +575,7 @@ wf::texture_t::texture_t(GLuint tex)
     this->tex_id = tex;
 }
 
-wf::texture_t::texture_t(wlr_texture *texture)
+wf::texture_t::texture_t(wlr_texture *texture, std::optional<wlr_fbox> viewport)
 {
     assert(wlr_texture_is_gles2(texture));
     wlr_gles2_texture_attribs attribs;
@@ -561,27 +594,20 @@ wf::texture_t::texture_t(wlr_texture *texture)
     {
         this->type = wf::TEXTURE_TYPE_EXTERNAL;
     }
-}
 
-wf::texture_t::texture_t(wlr_surface *surface) :
-    texture_t(surface->buffer->texture)
-{
-    if (surface->current.viewport.has_src)
+    if (viewport)
     {
         this->has_viewport = true;
 
-        auto width  = surface->buffer->texture->width;
-        auto height = surface->buffer->texture->height;
-
-        wlr_fbox fbox;
-        wlr_surface_get_buffer_source_box(surface, &fbox);
-        viewport_box.x1 = fbox.x / width;
-        viewport_box.x2 = (fbox.x + fbox.width) / width;
-        viewport_box.y1 = 1.0 - (fbox.y + fbox.height) / height;
-        viewport_box.y2 = 1.0 - (fbox.y) / height;
+        auto width  = texture->width;
+        auto height = texture->height;
+        viewport_box.x1 = viewport->x / width;
+        viewport_box.x2 = (viewport->x + viewport->width) / width;
+        viewport_box.y1 = 1.0 - (viewport->y + viewport->height) / height;
+        viewport_box.y2 = 1.0 - (viewport->y) / height;
     }
 }
-}
+} // namespace wf
 
 namespace OpenGL
 {
@@ -608,6 +634,11 @@ class program_t::impl
         uniforms[active_program_idx][name] =
             GL_CALL(glGetUniformLocation(id[active_program_idx], name.c_str()));
 
+        if (uniforms[active_program_idx][name] == -1)
+        {
+            LOGE("Uniform ", name, " not found in program");
+        }
+
         return uniforms[active_program_idx][name];
     }
 
