@@ -1,21 +1,18 @@
-#include "opengl-priv.hpp"
 #include "wayfire/output.hpp"
 #include "wayfire/core.hpp"
 #include "wayfire/output-layout.hpp"
-#include "wayfire/workspace-manager.hpp"
+#include "wayfire/view.hpp"
+#include "wayfire/workspace-set.hpp"
 #include "wayfire/render-manager.hpp"
 #include "wayfire/signal-definitions.hpp"
 #include "wayfire/util.hpp"
 
 #include "../output/output-impl.hpp"
-#include "seat/seat.hpp"
-#include "seat/cursor.hpp"
-#include "core-impl.hpp"
-
 #include <xf86drmMode.h>
-#include <sstream>
 #include <cstring>
 #include <unordered_set>
+#include <drm_fourcc.h>
+#include <wayfire/seat.hpp>
 
 #include <wayfire/debug.hpp>
 #include <wayfire/util/log.hpp>
@@ -141,81 +138,45 @@ void transfer_views(wf::output_t *from, wf::output_t *to)
 {
     assert(from);
 
-    LOGI("transfer views from ", from->handle->name, " -> ",
-        to ? to->handle->name : "null");
-    /* first move each desktop view(e.g windows) to another output */
-    std::vector<wayfire_view> views;
-    std::vector<wayfire_view> unmapped_views;
+    LOGI("transfer views from ", from->handle->name, " -> ", to ? to->handle->name : "null");
+
+    // Step 1: move views from the current workspace set to the other output
     if (to)
     {
         /* If we aren't moving to another output, then there is no need to
          * enumerate views either */
-        views = from->workspace->get_views_in_layer(wf::WM_LAYERS);
-
-        // Also get a list of views which are on that output, but do not have
-        // a layer. These are usually unmapped Xwayland views, which are not to
-        // be killed, as they are needed for the "real" views.
-        for (auto& view : wf::get_core().get_all_views())
-        {
-            if ((view->get_output() == from) &&
-                (from->workspace->get_view_layer(view) == 0) &&
-                (view->role != VIEW_ROLE_DESKTOP_ENVIRONMENT))
-            {
-                unmapped_views.push_back(view);
-            }
-        }
-
-        std::reverse(views.begin(), views.end());
-    }
-
-    for (auto& view : views)
-    {
-        from->workspace->remove_view(view);
-    }
-
-    /* views would be empty if !to, but clang-analyzer detects null deref */
-    if (to)
-    {
-        for (auto& view : unmapped_views)
-        {
-            // Most operations for transferring an unmapped view to another
-            // output don't make any sense, so we handle them separately.
-            view->set_output(to);
-        }
-
+        auto views = from->wset()->get_views(WSET_SORT_STACKING);
         for (auto& view : views)
         {
-            wf::get_core().move_view_to_output(view, to, true);
+            move_view_to_output(view, to, true);
         }
     }
 
-    // Find all leftover views
-    std::vector<wayfire_view> reffed;
+    // Step 2: Ensure none of the remaining views have an invalid output.
+    // Note that all views in workspace sets will have their output reassigned automatically by the
+    // workspace-set impl.
+    std::vector<std::shared_ptr<wf::view_interface_t>> non_ws_views;
     for (auto& view : wf::get_core().get_all_views())
     {
-        if (view->get_output() != from)
+        if ((view->get_output() == from) && (!toplevel_cast(view) || !toplevel_cast(view)->get_wset()))
         {
-            continue;
+            // Take a ref, so that the view doesn't get destroyed while we're doing operations on the views
+            non_ws_views.push_back(view->shared_from_this());
         }
-
-        // Ensure that no view is destroyed before we're finished with it!
-        // This is necessary in case we have for ex. a popup on a layer-shell
-        // surface. We don't know in which order they will be closed/destroyed.
-        reffed.push_back(view);
-        view->take_ref();
     }
 
-    // Close the leftover views, typically layer-shell ones
-    for (auto& view : reffed)
+    for (auto& view : non_ws_views)
     {
-        view->close();
-        view->set_output(nullptr);
-    }
-
-    // Drop refs we have taken
-    for (auto& view : reffed)
-    {
-        view->unref();
+        if (view->role == VIEW_ROLE_DESKTOP_ENVIRONMENT)
+        {
+            // typically: layer-shell views that should be closed, since they are tied to a single output
+            view->close();
+            view->set_output(nullptr);
+        } else
+        {
+            // typically: xwayland OR views
+            view->set_output(to);
+        }
     }
 }
 
@@ -241,6 +202,8 @@ bool output_state_t::operator ==(const output_state_t& other) const
     eq &= (mode.refresh == other.mode.refresh);
     eq &= (transform == other.transform);
     eq &= (scale == other.scale);
+    eq &= (vrr == other.vrr);
+    eq &= (depth == other.depth);
 
     return eq;
 }
@@ -250,114 +213,49 @@ inline bool is_shutting_down()
     return wf::get_core().get_current_state() == compositor_state_t::SHUTDOWN;
 }
 
-class output_cloner_t
+static const char *get_format_name(uint32_t format)
 {
-    wf::wl_listener_wrapper source_commit;
-    wf::wl_listener_wrapper destination_frame;
-    wf::framebuffer_base_t content;
-
-    wlr_output *source;
-    wlr_output *destination;
-
-  public:
-    output_cloner_t(wlr_output *source, wlr_output *destination)
+    switch (format)
     {
-        this->source = source;
-        this->destination = destination;
-        wlr_output_lock_software_cursors(source, true);
-        wlr_output_schedule_frame(destination);
+      case DRM_FORMAT_XRGB2101010:
+        return "DRM_FORMAT_XRGB2101010";
 
-        source_commit.set_callback([=] (void *data)
-        {
-            auto ev = (wlr_output_event_commit*)data;
-            if (!(ev->committed & WLR_OUTPUT_STATE_BUFFER))
-            {
-                // Something else than the output contents changed, nothing to do yet
-                return;
-            }
+      case DRM_FORMAT_XBGR2101010:
+        return "DRM_FORMAT_XBGR2101010";
 
-            int w = ev->buffer->width;
-            int h = ev->buffer->height;
+      case DRM_FORMAT_XRGB8888:
+        return "DRM_FORMAT_XRGB8888";
 
-            OpenGL::render_begin();
-            content.allocate(w, h);
-
-            // Bind buffer
-            auto renderer = get_core().renderer;
-            wlr_renderer_begin_with_buffer(renderer, ev->buffer);
-
-            // Store a copy for ourselves
-            GL_CALL(glBindFramebuffer(GL_DRAW_FRAMEBUFFER, content.fb));
-            GL_CALL(glBlitFramebuffer(0, 0, w, h,
-                0, 0, w, h,
-                GL_COLOR_BUFFER_BIT, GL_LINEAR));
-
-            wlr_renderer_end(renderer);
-            OpenGL::render_end();
-
-            wlr_output_damage_whole(destination);
-            wlr_output_schedule_frame(destination);
-        });
-
-        destination_frame.set_callback([=] (void *data)
-        {
-            auto renderer = get_core().renderer;
-            wlr_output_attach_render(destination, NULL);
-            wlr_renderer_begin(renderer, destination->width, destination->height);
-
-            int w = content.viewport_width;
-            int h = content.viewport_height;
-            if ((w > 0) && (h > 0))
-            {
-                int current_fb;
-                GL_CALL(glGetIntegerv(GL_DRAW_FRAMEBUFFER_BINDING, &current_fb));
-                OpenGL::bind_output(current_fb);
-
-                OpenGL::render_begin();
-                GL_CALL(glBindFramebuffer(GL_READ_FRAMEBUFFER, content.fb));
-                GL_CALL(glBlitFramebuffer(0, 0, w, h,
-                    0, 0, destination->width, destination->height,
-                    GL_COLOR_BUFFER_BIT, GL_LINEAR));
-
-                OpenGL::render_end();
-            }
-
-            wlr_renderer_end(renderer);
-            wlr_output_commit(destination);
-        });
-
-        source_commit.connect(&source->events.commit);
-        destination_frame.connect(&destination->events.frame);
+      case DRM_FORMAT_INVALID:
+      default:
+        return "DRM_FORMAT_INVALID";
     }
-
-    ~output_cloner_t()
-    {
-        wlr_output_lock_software_cursors(source, false);
-        OpenGL::render_begin();
-        content.release();
-        OpenGL::render_end();
-    }
-
-    output_cloner_t(const output_cloner_t&) = delete;
-    output_cloner_t(output_cloner_t&&) = delete;
-    output_cloner_t& operator =(const output_cloner_t&) = delete;
-    output_cloner_t& operator =(output_cloner_t&&) = delete;
-};
+}
 
 /** Represents a single output in the output layout */
 struct output_layout_output_t
 {
     wlr_output *handle;
     output_state_t current_state;
+    bool is_externally_managed = false;
+    bool is_nested_compositor  = false;
+    bool inhibited = false;
+    std::map<int, std::vector<uint32_t>> formats_for_depth;
+    int current_bit_depth = RENDER_BIT_DEPTH_DEFAULT;
 
     std::unique_ptr<wf::output_impl_t> output;
-    wl_listener_wrapper on_destroy, on_mode;
+    wl_listener_wrapper on_destroy, on_commit;
 
     std::shared_ptr<wf::config::section_t> config_section;
     wf::option_wrapper_t<wf::output_config::mode_t> mode_opt;
     wf::option_wrapper_t<wf::output_config::position_t> position_opt;
     wf::option_wrapper_t<double> scale_opt;
     wf::option_wrapper_t<std::string> transform_opt;
+    wf::option_wrapper_t<bool> vrr_opt;
+    wf::option_wrapper_t<int> depth_opt;
+
+    wf::option_wrapper_t<bool> use_ext_config{
+        "workarounds/use_external_output_configuration"};
 
     void initialize_config_options()
     {
@@ -367,6 +265,8 @@ struct output_layout_output_t
         position_opt.load_option(name + "/position");
         scale_opt.load_option(name + "/scale");
         transform_opt.load_option(name + "/transform");
+        vrr_opt.load_option(name + "/vrr");
+        depth_opt.load_option(name + "/depth");
     }
 
     output_layout_output_t(wlr_output *handle)
@@ -375,7 +275,7 @@ struct output_layout_output_t
         on_destroy.connect(&handle->events.destroy);
         initialize_config_options();
 
-        bool is_nested_compositor = wlr_output_is_wl(handle);
+        is_nested_compositor = wlr_output_is_wl(handle);
 
 #if WLR_HAS_X11_BACKEND
         is_nested_compositor |= wlr_output_is_x11(handle);
@@ -384,12 +284,23 @@ struct output_layout_output_t
         {
             /* Nested backends can be resized by the user. We need to handle
              * these cases */
-            on_mode.set_callback([=] (void *data)
+            on_commit.set_callback([=] (void *data)
             {
-                handle_mode_changed();
+                wlr_output_event_commit *ev = static_cast<wlr_output_event_commit*>(data);
+                if (ev->state->committed & WLR_OUTPUT_STATE_MODE)
+                {
+                    handle_mode_changed();
+                }
             });
-            on_mode.connect(&handle->events.mode);
+            on_commit.connect(&handle->events.commit);
         }
+
+        formats_for_depth[8]  = {DRM_FORMAT_XRGB8888};
+        formats_for_depth[10] = {
+            DRM_FORMAT_XRGB2101010,
+            DRM_FORMAT_XBGR2101010,
+            DRM_FORMAT_XRGB8888,
+        };
     }
 
     /**
@@ -412,6 +323,12 @@ struct output_layout_output_t
                 this->output->set_effective_size(get_effective_size());
                 this->output->render->damage_whole();
                 emit_configuration_changed(wf::OUTPUT_MODE_CHANGE);
+
+                // Emit the output-layout-configuration-changed signal as well, which is usually emitted for
+                // all changed outputs together in output-layout::apply_state(). However, resizing nested
+                // backends does not use apply_state(), hence we have to emit the signal manually.
+                output_layout_configuration_changed_signal ev;
+                wf::get_core().output_layout->emit(&ev);
             }
         }
     }
@@ -439,7 +356,7 @@ struct output_layout_output_t
         wlr_output_mode default_mode;
         auto width   = handle->width > 0 ? handle->width : 1200;
         auto height  = handle->height > 0 ? handle->height : 720;
-        auto refresh = handle->refresh > 0 ? handle->refresh : 60000;
+        auto refresh = handle->refresh > 0 ? handle->refresh : 0;
 
         default_mode.width   = width;
         default_mode.height  = height;
@@ -471,15 +388,50 @@ struct output_layout_output_t
         return true;
     }
 
-    output_state_t load_state_from_config()
+    /**
+     * Determine whether the state in the config file should be ignored.
+     */
+    bool should_ignore_config_state()
     {
+        if (is_externally_managed && use_ext_config)
+        {
+            wf::output_config::mode_t mode = mode_opt;
+            if (mode.get_type() == output_config::MODE_MIRROR)
+            {
+                // Special case: output mirroring
+                // It is not supported directly supported by wlr-output-management
+                // Thus, if the config file says to mirror an output, we do use that
+                // information.
+                return false;
+            }
+
+            return true;
+        }
+
+        return false;
+    }
+
+    /**
+     * Load the state the output is configured with.
+     * This is typically the config file, but in case of daemons like kanshi this
+     * might be the external configuration.
+     */
+    output_state_t load_configured_state()
+    {
+        // Ensure custom modes from the config are enabled
+        // Also make sure to refresh them even if the output is externally
+        // managed.
+        refresh_custom_modes();
+
+        if (should_ignore_config_state())
+        {
+            // Current state is what was requested by the client.
+            return this->current_state;
+        }
+
         output_state_t state;
         state.position = position_opt;
 
-        /* Make sure we can use custom modes that are
-         * specified in the config */
-        refresh_custom_modes();
-
         wf::output_config::mode_t mode = mode_opt;
         wlr_output_mode tmp;
 
@@ -515,6 +467,8 @@ struct output_layout_output_t
 
         state.scale     = scale_opt;
         state.transform = get_transform_from_string(transform_opt);
+        state.vrr   = vrr_opt;
+        state.depth = depth_opt;
         return state;
     }
 
@@ -534,25 +488,16 @@ struct output_layout_output_t
         /* Focus the first output, but do not change the focus on subsequently
          * added outputs. We also change the focus if the noop output was
          * focused */
-        wlr_output *focused = get_core().get_active_output() ?
-            get_core().get_active_output()->handle : nullptr;
+        wlr_output *focused = get_core().seat->get_active_output() ?
+            get_core().seat->get_active_output()->handle : nullptr;
         if (!focused || (focused->data == WF_NOOP_OUTPUT_MAGIC))
         {
-            get_core().focus_output(wo);
+            get_core().seat->focus_output(wo);
         }
 
-        /*
-         * At this point, this->output is a valid output and is part of the
-         * get_outputs() list.
-         *
-         * We have also have updated the focused output. So, at this point
-         * all plugin-relevant structures have been updated.
-         */
-        this->output->start_plugins();
-
         output_added_signal data;
         data.output = wo;
-        get_core().output_layout->emit_signal("output-added", &data);
+        get_core().output_layout->emit(&data);
     }
 
     void destroy_wayfire_output()
@@ -565,28 +510,30 @@ struct output_layout_output_t
         LOGE("disabling output: ", output->handle->name);
 
         auto wo = output.get();
+
         output_pre_remove_signal data;
         data.output = wo;
-
-        wo->emit_signal("pre-remove", &data);
-        get_core().output_layout->emit_signal("output-pre-remove", &data);
+        wo->emit(&data);
+        get_core().output_layout->emit(&data);
         wo->cancel_active_plugins();
 
         bool shutdown = is_shutting_down();
-        if ((get_core().get_active_output() == wo) && !shutdown)
+        if ((get_core().seat->get_active_output() == wo) && !shutdown)
         {
-            get_core().focus_output(
+            get_core().seat->focus_output(
                 get_core().output_layout->get_next_output(wo));
         } else if (shutdown)
         {
-            get_core().focus_output(nullptr);
+            get_core().seat->focus_output(nullptr);
         }
 
         /* It doesn't make sense to transfer to another output if we're
          * going to shut down the compositor */
-        transfer_views(wo,
-            shutdown ? nullptr : get_core().get_active_output());
-        get_core().output_layout->emit_signal("output-removed", &data);
+        transfer_views(wo, shutdown ? nullptr : get_core().seat->get_active_output());
+
+        wf::output_removed_signal data2;
+        data2.output = wo;
+        get_core().output_layout->emit(&data2);
         this->output = nullptr;
     }
 
@@ -632,20 +579,7 @@ struct output_layout_output_t
     /** Check whether the given state can be applied */
     bool test_state(const output_state_t& state)
     {
-        if (state.source == OUTPUT_IMAGE_SOURCE_NONE)
-        {
-            return true;
-        }
-
-        if (state.source == OUTPUT_IMAGE_SOURCE_MIRROR)
-        {
-            return true;
-        }
-
-        /* XXX: are there more things to check? */
-        refresh_custom_modes();
-
-        return is_mode_supported(state.mode);
+        return true;
     }
 
     /** Change the output mode */
@@ -656,7 +590,9 @@ struct output_layout_output_t
             /* Do not modeset if nothing changed */
             if ((handle->current_mode->width == mode.width) &&
                 (handle->current_mode->height == mode.height) &&
-                (handle->current_mode->refresh == mode.refresh))
+                (handle->current_mode->refresh == mode.refresh) &&
+                ((handle->adaptive_sync_status == WLR_OUTPUT_ADAPTIVE_SYNC_ENABLED) == current_state.vrr) &&
+                (current_bit_depth == current_state.depth))
             {
                 /* Commit the enabling of the output */
                 wlr_output_commit(handle);
@@ -677,11 +613,94 @@ struct output_layout_output_t
                 " for output ", handle->name, ". Trying to use custom mode",
                 "(might not work)");
 
-            wlr_output_set_custom_mode(handle, mode.width, mode.height,
-                mode.refresh);
+            wlr_output_set_custom_mode(handle, mode.width, mode.height, mode.refresh);
         }
 
         wlr_output_commit(handle);
+
+        const bool adaptive_sync_enabled = (handle->adaptive_sync_status == WLR_OUTPUT_ADAPTIVE_SYNC_ENABLED);
+
+        if (adaptive_sync_enabled != current_state.vrr)
+        {
+            wlr_output_enable_adaptive_sync(handle, current_state.vrr);
+            if (wlr_output_test(handle))
+            {
+                wlr_output_commit(handle);
+                LOGD("Changed adaptive sync on output: ", handle->name, " to ", current_state.vrr);
+            } else
+            {
+                LOGE("Failed to change adaptive sync on output: ", handle->name);
+                wlr_output_rollback(handle);
+            }
+        }
+
+        if (current_state.depth != current_bit_depth)
+        {
+            for (auto fmt : formats_for_depth[current_state.depth])
+            {
+                wlr_output_set_render_format(handle, fmt);
+                if (wlr_output_test(handle))
+                {
+                    wlr_output_commit(handle);
+                    current_bit_depth = current_state.depth;
+                    LOGD("Set output format to ", get_format_name(fmt), " on output ", handle->name);
+                    break;
+                }
+
+                LOGD("Failed to set output format ", get_format_name(fmt), " on output ", handle->name);
+            }
+        }
+    }
+
+    /* Mirroring implementation */
+    wl_listener_wrapper on_mirrored_frame;
+    wl_listener_wrapper on_frame;
+    wlr_output *locked_cursors_on = NULL;
+
+    /** Render the output using texture as source */
+    void render_output(wlr_texture *texture)
+    {
+        auto renderer = get_core().renderer;
+        wlr_output_attach_render(handle, NULL);
+        wlr_renderer_begin(renderer, handle->width, handle->height);
+
+        wf::texture_t tex{texture};
+        OpenGL::render_transformed_texture(tex, {-1, -1, 2, 2});
+
+        wlr_renderer_end(renderer);
+        wlr_output_commit(handle);
+    }
+
+    /* Load output contents and render them */
+    wlr_buffer *source_back_buffer = NULL;
+
+    void handle_frame()
+    {
+        auto wo = get_core().output_layout->find_output(
+            current_state.mirror_from);
+        if (!wo)
+        {
+            LOGE("Cannot find mirrored output ", current_state.mirror_from,
+                " for output ", handle->name);
+
+            return;
+        }
+
+        if (source_back_buffer == NULL)
+        {
+            LOGE("Got empty buffer on ", wo->handle->name);
+            return;
+        }
+
+        auto texture = wlr_texture_from_buffer(get_core().renderer, source_back_buffer);
+        if (!texture)
+        {
+            LOGE("Failed to export texture to dmabuf!");
+            return;
+        }
+
+        render_output(texture);
+        wlr_texture_destroy(texture);
     }
 
     void set_enabled(bool enabled)
@@ -693,7 +712,6 @@ struct output_layout_output_t
         }
     }
 
-    std::unique_ptr<output_cloner_t> cloner;
     void setup_mirror()
     {
         /* Check if we can mirror */
@@ -724,12 +742,58 @@ struct output_layout_output_t
             return;
         }
 
-        cloner = std::make_unique<output_cloner_t>(wo->handle, this->handle);
+        /* Force software cursors on the mirrored from output.
+         * This ensures that they will be copied when reading pixels
+         * from the main plane */
+        wlr_output_lock_software_cursors(wo->handle, true);
+        locked_cursors_on = wo->handle;
+
+        wlr_output_schedule_frame(handle);
+        on_mirrored_frame.set_callback([=] (void *data)
+        {
+            auto ev = (wlr_output_event_commit*)data;
+            if (!ev || !ev->state || !ev->state->buffer)
+            {
+                return;
+            }
+
+            if (ev->state->buffer)
+            {
+                if (source_back_buffer)
+                {
+                    wlr_buffer_unlock(source_back_buffer);
+                }
+
+                source_back_buffer = ev->state->buffer;
+                wlr_buffer_lock(ev->state->buffer);
+            }
+
+            /* The mirrored output was repainted, schedule repaint
+             * for us as well */
+            wlr_output_schedule_frame(handle);
+        });
+        on_mirrored_frame.connect(&wo->handle->events.commit);
+
+        on_frame.set_callback([=] (void*) { handle_frame(); });
+        on_frame.connect(&handle->events.frame);
     }
 
     void teardown_mirror()
     {
-        cloner.reset();
+        if (locked_cursors_on)
+        {
+            wlr_output_lock_software_cursors(locked_cursors_on, false);
+            locked_cursors_on = NULL;
+        }
+
+        if (source_back_buffer)
+        {
+            wlr_buffer_unlock(source_back_buffer);
+            source_back_buffer = NULL;
+        }
+
+        on_mirrored_frame.disconnect();
+        on_frame.disconnect();
     }
 
     wf::dimensions_t get_effective_size()
@@ -751,7 +815,7 @@ struct output_layout_output_t
             wf::output_configuration_changed_signal data{current_state};
             data.output = output.get();
             data.changed_fields = changed_fields;
-            output->emit_signal("output-configuration-changed", &data);
+            output->emit(&data);
         }
     }
 
@@ -811,6 +875,7 @@ struct output_layout_output_t
 
         set_enabled(!(state.source & OUTPUT_IMAGE_SOURCE_NONE));
         apply_mode(state.mode);
+
         if (state.source & OUTPUT_IMAGE_SOURCE_SELF)
         {
             if (handle->transform != state.transform)
@@ -821,8 +886,6 @@ struct output_layout_output_t
             if (handle->scale != state.scale)
             {
                 wlr_output_set_scale(handle, state.scale);
-                wf::get_core_impl().seat->cursor->load_xcursor_scale(
-                    state.scale);
             }
 
             wlr_output_commit(handle);
@@ -854,7 +917,7 @@ class output_layout_t::impl
     wl_listener_wrapper on_backend_destroy;
 
     wl_idle_call idle_update_configuration;
-    wl_timer timer_remove_noop;
+    wl_timer<false> timer_remove_noop;
 
     wlr_backend *noop_backend;
     /* Wayfire generally assumes that an enabled output is always available.
@@ -863,8 +926,13 @@ class output_layout_t::impl
      * virtual output with the noop backend. */
     std::unique_ptr<output_layout_output_t> noop_output;
 
-    signal_callback_t on_config_reload;
-    signal_connection_t on_backend_started = [=] (wf::signal_data_t*)
+    wf::signal::connection_t<wf::reload_config_signal> on_config_reload = [=] (wf::reload_config_signal *ev)
+    {
+        reconfigure_from_config();
+    };
+
+    wf::signal::connection_t<core_backend_started_signal> on_backend_started =
+        [=] (core_backend_started_signal *ev)
     {
         // We need to ensure that at any given time we have at least one
         // output while core is running.
@@ -905,14 +973,12 @@ class output_layout_t::impl
         on_backend_destroy.connect(&wf::get_core().renderer->events.destroy);
 
         output_layout = wlr_output_layout_create();
-
-        on_config_reload = [=] (void*) { reconfigure_from_config(); };
-        get_core().connect_signal("reload-config", &on_config_reload);
+        get_core().connect(&on_config_reload);
 
         noop_backend = wlr_headless_backend_create(get_core().display);
         wlr_backend_start(noop_backend);
 
-        get_core().connect_signal("_backend_started", &on_backend_started);
+        get_core().connect(&on_backend_started);
 
         output_manager = wlr_output_manager_v1_create(get_core().display);
         on_output_manager_test.set_callback([=] (void *data)
@@ -935,11 +1001,28 @@ class output_layout_t::impl
         on_output_power_mode_set.connect(&output_pw_manager->events.set_mode);
     }
 
-    ~impl()
+    void fini()
     {
-        get_core().disconnect_signal("reload-config", &on_config_reload);
+        // Destroy outputs first
+        this->outputs.clear();
+        noop_output.reset();
+
+        // Disconnect all signals
+        on_new_output.disconnect();
+        on_output_manager_test.disconnect();
+        on_output_manager_apply.disconnect();
+        on_output_power_mode_set.disconnect();
+        on_backend_destroy.disconnect();
+
+        wlr_backend_destroy(noop_backend);
+        wlr_output_layout_destroy(output_layout);
     }
 
+    impl(const impl &) = delete;
+    impl(impl &&) = delete;
+    impl& operator =(const impl&) = delete;
+    impl& operator =(impl&&) = delete;
+
     output_configuration_t output_configuration_from_wlr_configuration(
         wlr_output_configuration_v1 *configuration)
     {
@@ -964,11 +1047,29 @@ class output_layout_t::impl
             }
 
             state.source = OUTPUT_IMAGE_SOURCE_SELF;
-            state.mode   = head->state.mode ? *head->state.mode :
-                this->outputs[handle]->current_state.mode;
+
+            if (head->state.mode)
+            {
+                state.mode = *head->state.mode;
+            } else
+            {
+                state.mode.width   = head->state.custom_mode.width;
+                state.mode.height  = head->state.custom_mode.height;
+                state.mode.refresh = head->state.custom_mode.refresh;
+            }
+
             state.position  = {head->state.x, head->state.y};
             state.scale     = head->state.scale;
             state.transform = head->state.transform;
+            state.vrr = head->state.adaptive_sync_enabled;
+            if ((handle->pending.render_format == DRM_FORMAT_XRGB2101010) ||
+                (handle->pending.render_format == DRM_FORMAT_XBGR2101010))
+            {
+                state.depth = 10;
+            } else
+            {
+                state.depth = 8;
+            }
         }
 
         return result;
@@ -982,6 +1083,12 @@ class output_layout_t::impl
 
         if (apply_configuration(configuration, test_only))
         {
+            // Notify outputs that they have external configuration
+            for (auto& [wo, _] : configuration)
+            {
+                this->outputs[wo]->is_externally_managed = true;
+            }
+
             wlr_output_configuration_v1_send_succeeded(wlr_configuration);
         } else
         {
@@ -1014,7 +1121,7 @@ class output_layout_t::impl
          * next reconfiguration. This is needed because if we are removing
          * an output, we might get into a situation where the last physical
          * output has already been removed but we are yet to add the noop one */
-        noop_output->apply_state(noop_output->load_state_from_config());
+        noop_output->apply_state(noop_output->load_configured_state());
         wlr_output_layout_add_auto(output_layout, noop_output->handle);
         timer_remove_noop.disconnect();
     }
@@ -1037,11 +1144,26 @@ class output_layout_t::impl
         state.source = OUTPUT_IMAGE_SOURCE_NONE;
         noop_output->apply_state(state);
         wlr_output_layout_remove(output_layout, noop_output->handle);
+        // Trigger repositioning of all outputs
+        apply_configuration(get_current_configuration());
     }
 
     void add_output(wlr_output *output)
     {
-        LOGI("new output: ", output->name);
+        LOGI("new output: ", output->name,
+            " (\"", output->make, " ", output->model, " ", output->serial, "\")");
+
+        if (output->non_desktop)
+        {
+            LOGD("Non-desktop output ", output->name, " found");
+            if (get_core().protocols.drm_v1)
+            {
+                LOGD("Drm lease offered to ", output->name);
+                wlr_drm_lease_v1_manager_offer_output(get_core().protocols.drm_v1, output);
+            }
+
+            return;
+        }
 
         if (!wlr_output_init_render(output,
             get_core().allocator, get_core().renderer))
@@ -1088,27 +1210,22 @@ class output_layout_t::impl
         return configuration;
     }
 
-    output_configuration_t last_config_configuration;
-
     /** Load config from file, test and apply */
     void reconfigure_from_config()
     {
-        /* Load from config file */
+        // Load desired configuration from config file
         output_configuration_t configuration;
-        for (auto& entry : this->outputs)
+        for (auto& [output, layout_output] : this->outputs)
         {
-            configuration[entry.first] = entry.second->load_state_from_config();
+            configuration[output] = layout_output->load_configured_state();
         }
 
-        if ((configuration == get_current_configuration()) ||
-            (configuration == last_config_configuration))
+        if (configuration != get_current_configuration())
         {
-            return;
-        }
-
-        if (test_configuration(configuration))
-        {
-            apply_configuration(configuration);
+            if (test_configuration(configuration))
+            {
+                apply_configuration(configuration);
+            }
         }
     }
 
@@ -1373,15 +1490,6 @@ class output_layout_t::impl
             {
                 ++count_enabled;
                 wlr_output_layout_add_auto(output_layout, handle);
-
-                /* Get the correct position */
-                wlr_box box;
-                wlr_output_layout_get_box(output_layout, handle, &box);
-                if (wlr_box_empty(&box))
-                {
-                    LOGE("failed to get layout box");
-                }
-
                 lo->apply_state(state);
             }
         }
@@ -1400,7 +1508,23 @@ class output_layout_t::impl
             }
         }
 
-        get_core().output_layout->emit_signal("configuration-changed", nullptr);
+        /* Fifth: emit configuration-changed again for dynamically-positioned outputs, because their position
+         * might have changed. */
+        for (auto& entry : config)
+        {
+            auto& handle = entry.first;
+            auto& state  = entry.second;
+            auto& lo     = this->outputs[handle];
+
+            if (state.source & OUTPUT_IMAGE_SOURCE_SELF &&
+                entry.second.position.is_automatic_position())
+            {
+                lo->emit_configuration_changed(wf::OUTPUT_POSITION_CHANGE);
+            }
+        }
+
+        wf::output_layout_configuration_changed_signal ev;
+        get_core().output_layout->emit(&ev);
 
         if (count_enabled > 0)
         {
@@ -1412,7 +1536,6 @@ class output_layout_t::impl
             timer_remove_noop.set_timeout(1000, [=] ()
             {
                 remove_noop_output();
-                return false; // disconnect
             });
         }
 
@@ -1432,7 +1555,11 @@ class output_layout_t::impl
 
             wlr_box box;
             wlr_output_layout_get_box(output_layout, output.first, &box);
-            if (!wlr_box_empty(&box))
+            if (wlr_box_empty(&box))
+            {
+                head->state.x = 0;
+                head->state.y = 0;
+            } else
             {
                 head->state.x = box.x;
                 head->state.y = box.y;
@@ -1452,10 +1579,18 @@ class output_layout_t::impl
             return;
         }
 
-        config[ev->output].source =
-            (ev->mode == ZWLR_OUTPUT_POWER_V1_MODE_ON ?
-                OUTPUT_IMAGE_SOURCE_SELF : OUTPUT_IMAGE_SOURCE_DPMS);
+        const bool wants_dpms = (ev->mode == ZWLR_OUTPUT_POWER_V1_MODE_OFF);
+        config[ev->output].source = (wants_dpms ? OUTPUT_IMAGE_SOURCE_DPMS : OUTPUT_IMAGE_SOURCE_SELF);
         apply_configuration(config);
+
+        auto& wo = outputs[ev->output];
+        if (wo->inhibited != wants_dpms)
+        {
+            wo->inhibited = wants_dpms;
+            wo->output->render->add_inhibit(wants_dpms);
+        }
+
+        wo->output->render->damage_whole();
     }
 
     /* Public API functions */
@@ -1582,6 +1717,7 @@ class output_layout_t::impl
 output_layout_t::output_layout_t(wlr_backend *b) : pimpl(new impl(b))
 {}
 output_layout_t::~output_layout_t() = default;
+
 wlr_output_layout*output_layout_t::get_handle()
 {
     return pimpl->get_handle();
@@ -1633,4 +1769,9 @@ bool output_layout_t::apply_configuration(
 {
     return pimpl->apply_configuration(configuration, test_only);
 }
+
+void priv_output_layout_fini(wf::output_layout_t *layout)
+{
+    layout->pimpl->fini();
+}
 }
