@@ -1,14 +1,13 @@
-#include <cmath>
-
 #include <wayfire/util/log.hpp>
+#include <wayfire/bindings-repository.hpp>
 
 #include "touch.hpp"
 #include "cursor.hpp"
 #include "input-manager.hpp"
 #include "../core-impl.hpp"
+#include "wayfire/core.hpp"
 #include "wayfire/output.hpp"
-#include "wayfire/workspace-manager.hpp"
-#include "wayfire/compositor-surface.hpp"
+#include "wayfire/util.hpp"
 #include "wayfire/output-layout.hpp"
 
 wf::touch_interface_t::touch_interface_t(wlr_cursor *cursor, wlr_seat *seat,
@@ -22,76 +21,94 @@ wf::touch_interface_t::touch_interface_t(wlr_cursor *cursor, wlr_seat *seat,
     on_down.set_callback([=] (void *data)
     {
         auto ev   = static_cast<wlr_touch_down_event*>(data);
-        auto mode = emit_device_event_signal("touch_down", ev);
+        auto mode = emit_device_event_signal(ev, &ev->touch->base);
+        if (mode != input_event_processing_mode_t::IGNORE)
+        {
+            double lx, ly;
+            wlr_cursor_absolute_to_layout_coords(cursor, &ev->touch->base, ev->x, ev->y, &lx, &ly);
 
-        double lx, ly;
-        wlr_cursor_absolute_to_layout_coords(cursor, &ev->touch->base,
-            ev->x, ev->y, &lx, &ly);
+            wf::pointf_t point;
+            wf::get_core().output_layout->get_output_coords_at({lx, ly}, point);
+            handle_touch_down(ev->touch_id, ev->time_msec, point, mode);
+        }
 
-        wf::pointf_t point;
-        wf::get_core().output_layout->get_output_coords_at({lx, ly}, point);
-        handle_touch_down(ev->touch_id, ev->time_msec, point, mode);
-        wlr_idle_notify_activity(wf::get_core().protocols.idle,
-            wf::get_core().get_current_seat());
-        emit_device_event_signal("touch_down_post", ev);
+        wf::get_core().seat->notify_activity();
+        emit_device_post_event_signal(ev, &ev->touch->base);
     });
 
     on_up.set_callback([=] (void *data)
     {
         auto ev   = static_cast<wlr_touch_up_event*>(data);
-        auto mode = emit_device_event_signal("touch_up", ev);
-        handle_touch_up(ev->touch_id, ev->time_msec, mode);
-        wlr_idle_notify_activity(wf::get_core().protocols.idle,
-            wf::get_core().get_current_seat());
-        emit_device_event_signal("touch_up_post", ev);
+        auto mode = emit_device_event_signal(ev, &ev->touch->base);
+        if (mode != input_event_processing_mode_t::IGNORE)
+        {
+            handle_touch_up(ev->touch_id, ev->time_msec, mode);
+        }
+
+        wf::get_core().seat->notify_activity();
+        emit_device_post_event_signal(ev, &ev->touch->base);
     });
 
     on_motion.set_callback([=] (void *data)
     {
         auto ev   = static_cast<wlr_touch_motion_event*>(data);
-        auto mode = emit_device_event_signal("touch_motion", ev);
+        auto mode = emit_device_event_signal(ev, &ev->touch->base);
 
-        double lx, ly;
-        wlr_cursor_absolute_to_layout_coords(
-            wf::get_core_impl().seat->cursor->cursor, &ev->touch->base,
-            ev->x, ev->y, &lx, &ly);
+        if (mode != input_event_processing_mode_t::IGNORE)
+        {
+            double lx, ly;
+            wlr_cursor_absolute_to_layout_coords(
+                wf::get_core_impl().seat->priv->cursor->cursor, &ev->touch->base,
+                ev->x, ev->y, &lx, &ly);
 
-        wf::pointf_t point;
-        wf::get_core().output_layout->get_output_coords_at({lx, ly}, point);
-        handle_touch_motion(ev->touch_id, ev->time_msec, point, true, mode);
-        wlr_idle_notify_activity(wf::get_core().protocols.idle,
-            wf::get_core().get_current_seat());
-        emit_device_event_signal("touch_motion_post", ev);
+            wf::pointf_t point;
+            wf::get_core().output_layout->get_output_coords_at({lx, ly}, point);
+            handle_touch_motion(ev->touch_id, ev->time_msec, point, true, mode);
+        }
+
+        wf::get_core().seat->notify_activity();
+        emit_device_post_event_signal(ev, &ev->touch->base);
+    });
+
+    on_cancel.set_callback([=] (void *data)
+    {
+        wlr_touch_cancel_event *ev = (wlr_touch_cancel_event*)data;
+        wlr_touch_up_event sim;
+        sim.time_msec = ev->time_msec;
+        sim.touch_id  = ev->touch_id;
+        sim.touch     = ev->touch;
+        this->on_up.emit(&sim);
+    });
+
+    on_frame.set_callback([&] (void*)
+    {
+        wlr_seat_touch_notify_frame(wf::get_core().get_current_seat());
+        wf::get_core().seat->notify_activity();
     });
 
     on_up.connect(&cursor->events.touch_up);
     on_down.connect(&cursor->events.touch_down);
+    on_frame.connect(&cursor->events.touch_frame);
     on_motion.connect(&cursor->events.touch_motion);
+    on_cancel.connect(&cursor->events.touch_cancel);
 
-    on_surface_map_state_change.set_callback(
-        [=] (wf::surface_interface_t *surface)
+    on_root_node_updated = [=] (wf::scene::root_node_update_signal *data)
     {
-        if ((this->grabbed_surface == surface) && !surface->is_mapped())
+        if (!(data->flags & wf::scene::update_flag::INPUT_STATE))
         {
-            end_touch_down_grab();
-            on_stack_order_changed.emit(nullptr);
+            return;
         }
-    });
 
-    on_stack_order_changed.set_callback([=] (wf::signal_data_t *data)
-    {
-        for (auto f : this->get_state().fingers)
+        for (auto& [finger_id, node] : this->focus)
         {
-            this->handle_touch_motion(f.first, get_current_time(),
-                {f.second.current.x, f.second.current.y}, false,
-                input_event_processing_mode_t::FULL);
+            if (node && !is_grabbed_node_alive(node))
+            {
+                set_touch_focus(nullptr, finger_id, get_current_time(), {0, 0});
+            }
         }
-    });
-
-    wf::get_core().connect_signal("output-stack-order-changed",
-        &on_stack_order_changed);
-    wf::get_core().connect_signal("view-geometry-changed", &on_stack_order_changed);
+    };
 
+    wf::get_core().scene()->connect(&on_root_node_updated);
     add_default_gestures();
 }
 
@@ -103,31 +120,10 @@ const wf::touch::gesture_state_t& wf::touch_interface_t::get_state() const
     return this->finger_state;
 }
 
-wf::surface_interface_t*wf::touch_interface_t::get_focus() const
+wf::scene::node_ptr wf::touch_interface_t::get_focus(int finger_id) const
 {
-    return this->focus;
-}
-
-void wf::touch_interface_t::set_grab(wf::plugin_grab_interface_t *grab)
-{
-    if (grab)
-    {
-        this->grab = grab;
-        end_touch_down_grab();
-        for (auto& f : this->get_state().fingers)
-        {
-            set_touch_focus(nullptr, f.first, get_current_time(), {0, 0});
-        }
-    } else
-    {
-        this->grab = nullptr;
-        for (auto& f : this->get_state().fingers)
-        {
-            handle_touch_motion(f.first, get_current_time(),
-                {f.second.current.x, f.second.current.y}, false,
-                input_event_processing_mode_t::FULL);
-        }
-    }
+    auto it = this->focus.find(finger_id);
+    return (it == focus.end() ? nullptr : it->second);
 }
 
 void wf::touch_interface_t::add_touch_gesture(
@@ -143,54 +139,40 @@ void wf::touch_interface_t::rem_touch_gesture(
         gestures.end());
 }
 
-void wf::touch_interface_t::set_touch_focus(wf::surface_interface_t *surface,
-    int id, uint32_t time, wf::pointf_t point)
+void wf::touch_interface_t::set_touch_focus(wf::scene::node_ptr node,
+    int id, int64_t time, wf::pointf_t point)
 {
-    bool focus_compositor_surface = wf::compositor_surface_from_surface(surface);
-    bool had_focus = wlr_seat_touch_get_point(seat, id);
-    wf::get_core_impl().seat->ensure_input_surface(surface);
-
-    wlr_surface *next_focus = NULL;
-    if (surface && !focus_compositor_surface)
+    if (focus[id] == node)
     {
-        next_focus = surface->get_wlr_surface();
+        return;
     }
 
-    // create a new touch point, we have a valid new focus
-    if (!had_focus && next_focus)
+    if (focus[id])
     {
-        wlr_seat_touch_notify_down(seat, next_focus, time, id, point.x, point.y);
+        focus[id]->touch_interaction().handle_touch_up(time, id, point);
     }
 
-    if (had_focus && !next_focus)
+    focus[id] = node;
+    if (node)
     {
-        wlr_seat_touch_notify_up(seat, time, id);
+        auto local = get_node_local_coords(node.get(), point);
+        node->touch_interaction().handle_touch_down(time, id, local);
     }
+}
 
-    if (next_focus)
+void wf::touch_interface_t::transfer_grab(scene::node_ptr grab_node)
+{
+    auto new_focus = grab_node->wants_raw_input() ? grab_node : nullptr;
+    for (auto& [id, focused_node] : this->focus)
     {
-        wlr_seat_touch_point_focus(seat, next_focus, time, id, point.x, point.y);
-    }
-
-    /* Manage the touch_focus, we take only the first finger for that */
-    if (id == 0)
-    {
-        // first change the focus, so that plugins can freely grab input in
-        // response to touch_down/up
-        auto old_focus = this->focus;
-        this->focus = surface;
-
-        auto compositor_surface = compositor_surface_from_surface(old_focus);
-        if (compositor_surface)
+        if (focused_node && (focused_node != new_focus) && !focused_node->wants_raw_input())
         {
-            compositor_surface->on_touch_up();
+            const auto lift_off_position = finger_state.fingers[id].current;
+            focused_node->touch_interaction().handle_touch_up(get_current_time(), id,
+                {lift_off_position.x, lift_off_position.y});
         }
 
-        compositor_surface = compositor_surface_from_surface(surface);
-        if (compositor_surface)
-        {
-            compositor_surface->on_touch_down(point.x, point.y);
-        }
+        focused_node = new_focus;
     }
 }
 
@@ -212,11 +194,11 @@ void wf::touch_interface_t::handle_touch_down(int32_t id, uint32_t time,
     wf::pointf_t point, input_event_processing_mode_t mode)
 {
     auto& seat = wf::get_core_impl().seat;
-    seat->break_mod_bindings();
+    seat->priv->break_mod_bindings();
 
     if (id == 0)
     {
-        wf::get_core().focus_output(
+        wf::get_core().seat->focus_output(
             wf::get_core().output_layout->get_output_at(point.x, point.y));
     }
 
@@ -233,34 +215,17 @@ void wf::touch_interface_t::handle_touch_down(int32_t id, uint32_t time,
     };
     finger_state.update(gesture_event);
 
-    if (this->grab || (mode != input_event_processing_mode_t::FULL))
+    if (mode != input_event_processing_mode_t::FULL)
     {
         update_gestures(gesture_event);
         update_cursor_state();
-        if (grab->callbacks.touch.down)
-        {
-            auto wo = wf::get_core().get_active_output();
-            auto og = wo->get_layout_geometry();
-            grab->callbacks.touch.down(id, point.x - og.x, point.y - og.y);
-        }
-
         return;
     }
 
-    wf::pointf_t local;
-    auto focus = this->surface_at(point, local);
-    if (finger_state.fingers.empty()) // finger state is not updated yet
-    {
-        start_touch_down_grab(focus);
-    } else if (grabbed_surface && !seat->drag_active)
-    {
-        focus = grabbed_surface;
-        local = get_surface_relative_coords(focus, point);
-    }
+    auto focus = this->surface_at(point);
+    set_touch_focus(focus, id, time, point);
 
-    set_touch_focus(focus, id, time, local);
-
-    seat->update_drag_icon();
+    seat->priv->update_drag_icon();
     update_gestures(gesture_event);
     update_cursor_state();
 }
@@ -286,99 +251,40 @@ void wf::touch_interface_t::handle_touch_motion(int32_t id, uint32_t time,
         finger_state.update(gesture_event);
     }
 
-    if (this->grab)
+    if (focus[id])
     {
-        auto wo = wf::get_core().output_layout->get_output_at(point.x, point.y);
-        auto og = wo->get_layout_geometry();
-        if (grab->callbacks.touch.motion && is_real_event)
-        {
-            grab->callbacks.touch.motion(id, point.x - og.x, point.y - og.y);
-        }
-
-        return;
+        auto local = get_node_local_coords(focus[id].get(), point);
+        focus[id]->touch_interaction().handle_touch_motion(time, id, local);
     }
 
-    wf::pointf_t local;
-    wf::surface_interface_t *surface = nullptr;
     auto& seat = wf::get_core_impl().seat;
-    /* Same as cursor motion handling: make sure we send to the grabbed surface,
-     * except if we need this for DnD */
-    if (grabbed_surface && !seat->drag_icon)
-    {
-        surface = grabbed_surface;
-        local   = get_surface_relative_coords(surface, point);
-    } else
-    {
-        surface = surface_at(point, local);
-        set_touch_focus(surface, id, time, local);
-    }
-
-    wlr_seat_touch_notify_motion(seat->seat, time, id, local.x, local.y);
-    seat->update_drag_icon();
-
-    auto compositor_surface = wf::compositor_surface_from_surface(surface);
-    if ((id == 0) && compositor_surface && is_real_event)
-    {
-        compositor_surface->on_touch_motion(local.x, local.y);
-    }
+    seat->priv->update_drag_icon();
 }
 
 void wf::touch_interface_t::handle_touch_up(int32_t id, uint32_t time,
     input_event_processing_mode_t mode)
 {
+    const auto lift_off_position = finger_state.fingers[id].current;
+
     const wf::touch::gesture_event_t gesture_event = {
         .type   = wf::touch::EVENT_TYPE_TOUCH_UP,
         .time   = time,
         .finger = id,
-        .pos    = finger_state.fingers[id].current
+        .pos    = lift_off_position,
     };
+
     update_gestures(gesture_event);
     finger_state.update(gesture_event);
 
     update_cursor_state();
-
-    if (this->grab)
-    {
-        if (grab->callbacks.touch.up)
-        {
-            grab->callbacks.touch.up(id);
-        }
-
-        return;
-    }
-
-    set_touch_focus(nullptr, id, time, {0, 0});
-    if (finger_state.fingers.empty())
-    {
-        end_touch_down_grab();
-    }
-}
-
-void wf::touch_interface_t::start_touch_down_grab(
-    wf::surface_interface_t *surface)
-{
-    this->grabbed_surface = surface;
-}
-
-void wf::touch_interface_t::end_touch_down_grab()
-{
-    if (grabbed_surface)
-    {
-        grabbed_surface = nullptr;
-        for (auto& f : finger_state.fingers)
-        {
-            handle_touch_motion(f.first, wf::get_current_time(),
-                {f.second.current.x, f.second.current.y}, false,
-                input_event_processing_mode_t::FULL);
-        }
-    }
+    set_touch_focus(nullptr, id, time, {lift_off_position.x, lift_off_position.y});
 }
 
 void wf::touch_interface_t::update_cursor_state()
 {
     auto& seat = wf::get_core_impl().seat;
     /* just set the cursor mode, independent of how many fingers we have */
-    seat->cursor->set_touchscreen_mode(true);
+    seat->priv->cursor->set_touchscreen_mode(true);
 }
 
 // Swipe params
@@ -498,7 +404,7 @@ class multi_action_t : public gesture_action_t
 
 static uint32_t find_swipe_edges(wf::touch::point_t point)
 {
-    auto output   = wf::get_core().get_active_output();
+    auto output   = wf::get_core().seat->get_active_output();
     auto geometry = output->get_layout_geometry();
 
     uint32_t edge_directions = 0;
@@ -603,13 +509,12 @@ void wf::touch_interface_t::add_default_gestures()
             direction,
             swp_ptr->cnt_fingers
         };
-        wf::get_core_impl().input->get_active_bindings().handle_gesture(gesture);
+        wf::get_core().bindings->handle_gesture(gesture);
     };
 
     auto ack_edge_swipe = [esw_ptr, this] ()
     {
-        uint32_t possible_edges =
-            find_swipe_edges(finger_state.get_center().origin);
+        uint32_t possible_edges = find_swipe_edges(finger_state.get_center().origin);
         uint32_t direction = wf_touch_to_wf_dir(esw_ptr->target_direction);
 
         possible_edges &= direction;
@@ -621,7 +526,7 @@ void wf::touch_interface_t::add_default_gestures()
                 esw_ptr->cnt_fingers
             };
 
-            wf::get_core_impl().input->get_active_bindings().handle_gesture(gesture);
+            wf::get_core().bindings->handle_gesture(gesture);
         }
     };
 
@@ -633,7 +538,7 @@ void wf::touch_interface_t::add_default_gestures()
             pnc_ptr->cnt_fingers
         };
 
-        wf::get_core_impl().input->get_active_bindings().handle_gesture(gesture);
+        wf::get_core().bindings->handle_gesture(gesture);
     };
 
     this->multiswipe = std::make_unique<gesture_t>(std::move(
